from .base import Parser


class BinanceParser(Parser):
    @staticmethod
    def check_response(response: dict):
        return {"code": 200, "status": "success", "data": response}

    @property
    def spot_exchange_info_parser(self):
        return {
            "active": (lambda x: x["status"] == "TRADING"),
            "is_spot": True,
            "is_margin": (lambda x: x["isMarginTradingAllowed"]),
            "is_futures": False,
            "is_perp": False,
            "is_linear": True,
            "is_inverse": False,
            "symbol": (lambda x: self.parse_unified_symbol(x["baseAsset"], x["quoteAsset"])),
            "base": (lambda x: str(x["baseAsset"])),
            "quote": (lambda x: str(x["quoteAsset"])),
            "settle": (lambda x: str(x["quoteAsset"])),
            "multiplier": 1,  # spot multiplier default 1
            "leverage": 1,  # spot leverage default 1
            "listing_time": None,  # api not support this field
            "expiration_time": None,  # spot not support this field
            "contract_size": 1,  # spot contract size default 1
            "tick_size": None,  # Not yet implemented
            "min_order_size": None,  # Not yet implemented
            "max_order_size": None,  # Not yet implemented
            "raw_data": (lambda x: x),
        }

    def futures_exchange_info_parser(self, market_type: str):
        return {
            "active": (lambda x: (x["status"] if market_type != "inverse" else x["contractStatus"]) == "TRADING"),
            "is_spot": False,
            "is_margin": False,
            "is_futures": (lambda x: self.parse_is_futures(x["contractType"])),
            "is_perp": (lambda x: self.parse_is_perpetual(x["contractType"])),
            "is_linear": True if market_type == "linear" else False,
            "is_inverse": True if market_type == "inverse" else False,
            "symbol": (lambda x: self.parse_unified_symbol(self.parse_base_currency(x["baseAsset"]), x["quoteAsset"])),
            "base": (lambda x: self.parse_base_currency(x["baseAsset"])),
            "quote": (lambda x: x["quoteAsset"]),
            "settle": (lambda x: x["marginAsset"]),
            "multiplier": (lambda x: self.parse_multiplier(x["baseAsset"])),
            "leverage": 1,  # need to find another way to get the leverage data
            "listing_time": (lambda x: int(x["onboardDate"])),
            "expiration_time": (lambda x: int(x["deliveryDate"])),
            "contract_size": (
                lambda x: 1 if "contractSize" not in x else float(x["contractSize"])
            ),  # binance only have contract size to inverse perp and futures
            "tick_size": None,  # not yet implemented
            "min_order_size": None,  # not yet implemented
            "max_order_size": None,  # not yet implemented
            "raw_data": (lambda x: x),
        }

    def parse_exchange_info(self, response: dict, parser: dict) -> dict:
        response = self.check_response(response)
        if response["code"] != 200:
            return response

        datas = response["data"]["symbols"]
        results = {}
        for data in datas:
            result = self.get_result_with_parser(data, parser)
            id = self.parse_unified_id(result)
            results[id] = result

        return results

    def parse_ticker(self, response: dict, info: dict) -> dict:
        if isinstance(response, list):
            response = response[0]

        base_volume = float(response["volume"] if info["is_linear"] else response["baseVolume"])
        quote_volume = float(response["quoteVolume"] if info["is_linear"] else response["volume"])

        quote_volume *= info["contract_size"] if info["is_perp"] or info["is_futures"] else 1

        return {
            "timestamp": self.parse_str(response["closeTime"], int),
            "instrument_id": self.parse_unified_id(info),
            "open_time": self.parse_str(response["openTime"], int),
            "close_time": self.parse_str(response["closeTime"], int),
            "open": self.parse_str(response["openPrice"], float),
            "high": self.parse_str(response["highPrice"], float),
            "low": self.parse_str(response["lowPrice"], float),
            "last": self.parse_str(response["lastPrice"], float),
            "base_volume": base_volume,
            "quote_volume": quote_volume,
            "price_change": self.parse_str(response["priceChange"], float),
            "price_change_percent": self.parse_str(response["priceChangePercent"], float) / 100,
            "raw_data": response,
        }

    def get_id_map(self, infos: dict, market_type: str) -> dict:
        infos = self.query_dict(infos, {f"is_{market_type}": True})
        return {v["raw_data"]["symbol"]: k for k, v in infos.items()}

    def parse_tickers(self, response: dict, market_type: str, infos: dict) -> dict:
        response = self.check_response(response)
        if response["code"] != 200:
            return response

        datas = response["data"]
        id_map = self.get_id_map(infos, market_type)
        results = {}
        for data in datas:
            if data["symbol"] not in id_map:
                print(data["symbol"])
                continue
            instrument_id = id_map[data["symbol"]]
            results[instrument_id] = self.parse_ticker(data, infos[instrument_id])
        return results

    def parse_spot_account_info(self, response: dict) -> dict:
        response = self.check_response(response)
        data = response["data"]
        return {
            "timestamp": data["updateTime"],
            "account_id": data["uid"],
            "account_type": data["accountType"],
            "raw_data": data,
        }

    def parse_margin_account_info(self, response: dict) -> dict:
        response = self.check_response(response)

        data = response["data"]
        return {"raw_data": data}

    def parse_margin_balance(self, response: dict, currency: str = None) -> dict:
        response = self.check_response(response)
        datas = response["data"]["userAssets"]
        query_currency = currency

        results = {}
        for data in datas:
            if data["netAsset"] == "0":
                continue

            result = {
                "currency": data["asset"],
                "balance": self.parse_str(data["netAsset"], float),
                "available": self.parse_str(data["free"], float),
                "raw_data": data,
            }
            currency = result["currency"]
            results[currency] = result
        return results if not query_currency else {query_currency: results[query_currency]}

    def parse_history_funding_rate(self, response: dict, info: dict) -> list:
        response = self.check_response(response)
        datas = response["data"]

        results = []
        for data in datas:
            result = {
                "timestamp": int(round(self.parse_str(data["fundingTime"], int) / 1000)) * 1000,
                "instrument_id": self.parse_unified_id(info),
                "market_type": self.parse_unified_market_type(info),
                "funding_rate": self.parse_str(data["fundingRate"], float),
                "realized_rate": None,
                "raw_data": data,
            }
            results.append(result)
        return results

    def parse_last_price(self, response: dict, instrument_id: str) -> dict:
        response = self.check_response(response)
        data = response["data"][instrument_id]

        return {
            "timestamp": data["close_time"],
            "instrument_id": instrument_id,
            "last_price": self.parse_str(data["last_price"], float),
            "raw_data": data,
        }

    def parse_index_price(self, response: dict, info: dict, market_type: str) -> dict:
        response = self.check_response(response)
        data = response["data"]

        if not isinstance(data, dict):
            data = data[0]

        if market_type == "spot":
            return {
                "timestamp": self.parse_str(data["calcTime"], int),
                "instrument_id": self.parse_unified_id(info),
                "index_price": self.parse_str(data["price"], float),
                "raw_data": data,
            }
        else:  # linear, inverse
            return {
                "timestamp": self.parse_str(data["time"], int),
                "instrument_id": self.parse_unified_id(info),
                "index_price": self.parse_str(data["indexPrice"], float),
                "raw_data": data,
            }

    def parse_mark_price(self, response: dict, info: dict, market_type: str) -> dict:
        response = self.check_response(response)
        data = response["data"]

        if not isinstance(data, dict):
            data = data[0]

        return {
            "timestamp": self.parse_str(data["time"], int),
            "instrument_id": self.parse_unified_id(info),
            "mark_price": self.parse_str(data["markPrice"], float),
            "raw_data": data,
        }

    def parse_open_interest(self, response: dict, info: dict, market_type: str) -> dict:
        response = self.check_response(response)
        data = response["data"]

        return {
            "timestamp": self.parse_str(data["time"], int),
            "instrument_id": self.parse_unified_id(info),
            "market_type": self.parse_unified_market_type(info),
            "oi_contract": self.parse_str(data["openInterest"], float),
            "oi_currency": None,
            "raw_data": data,
        }

    def parse_orderbook(self, response: dict, info: dict, market_type: str, depth: int) -> dict:
        response = self.check_response(response)
        data = response["data"]

        results = {
            "timestamp": self.get_timestamp(),
            "instrument_id": self.parse_unified_id(info),
            "asks": [
                {
                    "price": self.parse_str(ask[0], float),
                    "volume": self.parse_str(ask[1], float),
                    "order_number": None,
                }
                for ask in data["asks"]
            ],
            "bids": [
                {
                    "price": self.parse_str(bid[0], float),
                    "volume": self.parse_str(bid[1], float),
                    "order_number": None,
                }
                for bid in data["bids"]
            ],
            "raw_data": data,
        }

        results["bids"] = sorted(results["bids"], key=lambda x: x["price"], reverse=True)
        results["asks"] = sorted(results["asks"], key=lambda x: x["price"])

        if depth:
            results["bids"] = results["bids"][:depth]
            results["asks"] = results["asks"][:depth]

        return results

    def get_symbol(self, info: dict) -> str:
        return f'{info["base"]}{info["quote"]}'

    def get_interval(self, interval: str) -> str:
        return interval

    def get_market_type(self, info: dict):
        if info["is_spot"]:
            return "spot"
        elif info["is_linear"]:
            return "linear"
        elif info["is_inverse"]:
            return "inverse"
        else:
            raise Exception("market type not found")

    def parse_current_funding_rate(self, response: dict, info: dict) -> dict:
        response = self.check_response(response)
        data = response["data"]

        return {
            "timestamp": self.parse_str(data["time"], int),
            "next_funding_time": self.parse_str(data["nextFundingTime"], int),
            "instrument_id": self.parse_unified_id(info),
            "market_type": self.parse_unified_market_type(info),
            "funding_rate": self.parse_str(data["lastFundingRate"], float),
            "raw_data": data,
        }

    def parse_candlesticks(self, response: dict, info: dict, market_type: str, interval: str) -> any:
        response = self.check_response(response)
        datas = response["data"]
        instrument_id = self.parse_unified_id(info)
        market_type = self.parse_unified_market_type(info)

        results = []

        for data in datas:
            results.append(
                {
                    "timestamp": self.parse_str(data[0], int),
                    "instrument_id": instrument_id,
                    "market_type": market_type,
                    "interval": interval,
                    "open": self.parse_str(data[1], float),
                    "high": self.parse_str(data[2], float),
                    "low": self.parse_str(data[3], float),
                    "close": self.parse_str(data[4], float),
                    "base_volume": self.parse_str(data[5], float),
                    "quote_volume": self.parse_str(data[7], float),
                    "contract_volume": self.parse_str(data[5], float),
                    "raw_data": data,
                }
            )
        return results[0] if len(results) == 1 else results



from .base import BaseClient


class KucoinSpot(BaseClient):
    BASE_ENDPOINT = "https://api.kucoin.com"

    def __init__(self) -> None:
        super().__init__()
        self.spot_base_endpoint = self.BASE_ENDPOINT

    async def _get_currency_list(self):
        return await self._get(self.spot_base_endpoint + "/api/v3/currencies")

    async def _get_symbol_list(self):
        return await self._get(self.spot_base_endpoint + "/api/v2/symbols")

    async def _get_tickers(self):
        return await self._get(self.spot_base_endpoint + "/api/v1/market/allTickers")

    async def _get_24hr_ticker(self, symbol: str):
        params = {"symbol": symbol}
        return await self._get(self.spot_base_endpoint + "/api/v1/market/stats", params=params)

    async def _get_24hr_stats(self, symbol: str):
        return await self._get(self.spot_base_endpoint + f"/api/v1/market/stats?symbol={symbol}")

    async def _get_klines(self, symbol: str, type: str, start: int = None, end: int = None):
        params = {
            "symbol": symbol,
            "type": type,
        }
        if start:
            params["startAt"] = start

        if end:
            params["endAt"] = end

        return await self._get(self.spot_base_endpoint + "/api/v1/market/candles", params=params)

    async def _get_margin_mark_price(self, symbol: str):
        return await self._get(self.spot_base_endpoint + f"/api/v1/mark-price/{symbol}/current")

    async def _get_part_orderbook(self, symbol: str, depth: int):
        if depth not in [20, 100]:
            raise ValueError("Depth must be 20 or 100 in KucoinSpot")
        return await self._get(
            self.spot_base_endpoint + f"/api/v1/market/orderbook/level2_{depth}", params={"symbol": symbol}
        )

    async def _get_full_orderbook(self, symbol: str):
        return await self._get(self.spot_base_endpoint + f"/api/v3/market/orderbook/level2", params={"symbol": symbol})


class KucoinFutures(BaseClient):
    BASE_ENDPOINT = "https://api-futures.kucoin.com"

    def __init__(self) -> None:
        super().__init__()
        self.futures_base_endpoint = self.BASE_ENDPOINT

    async def _get_symbol_list(self):
        return await self._get(self.futures_base_endpoint + "/api/v1/contracts/active")

    async def _get_ticker(self, symbol: str):
        params = {"symbol": symbol}
        return await self._get(self.futures_base_endpoint + f"/api/v1/ticker", params=params)

    async def _get_symbol_detail(self, symbol: str):
        return await self._get(self.futures_base_endpoint + f"/api/v1/contracts/{symbol}")

    async def _get_klines(self, symbol: str, granularity: int, start: int = None, end: int = None):

        params = {
            "symbol": symbol,
            "granularity": granularity,
        }
        if start:
            params["from"] = start
        if end:
            params["to"] = end

        return await self._get(self.futures_base_endpoint + "/api/v1/kline/query", params=params)

    async def _get_current_mark_price(self, symbol: str):
        return await self._get(self.futures_base_endpoint + f"/api/v1/mark-price/{symbol}/current")

    async def _get_part_orderbook(self, symbol: str, depth: int):
        if depth not in [20, 100]:
            raise ValueError("Depth must be 20 or 100 in KucoinSpot")
        return await self._get(self.futures_base_endpoint + f"/api/v1/level2/depth{depth}", params={"symbol": symbol})

    async def _get_full_orderbook(self, symbol: str):
        return await self._get(self.futures_base_endpoint + "/api/v1/level2/snapshot", params={"symbol": symbol})

    async def _get_current_funding_rate(self, symbol: str):
        return await self._get(self.futures_base_endpoint + f"/api/v1/funding-rate/{symbol}/current")

    async def _get_public_funding_history(self, symbol: str, _from: int, to: int):
        params = {
            "symbol": symbol,
            "from": _from,
            "to": to,
        }
        return await self._get(self.futures_base_endpoint + "/api/v1/contract/funding-rates", params=params)

    async def _get_private_funding_history(
        self,
        symbol: str,
        startAt: int = None,
        endAt: int = None,
        reverse: bool = None,
        offset: int = None,
        forward: bool = None,
        maxCount: int = None,
    ):
        params = {
            k: v
            for k, v in {
                "symbol": symbol,
                "startAt": startAt,
                "endAt": endAt,
                "reverse": reverse,
                "offset": offset,
                "forward": forward,
                "maxCount": maxCount,
            }.items()
            if v is not None
        }
        return await self._get(self.futures_base_endpoint + "/api/v1/funding-history", params=params)



from .base import Parser


class GateioParser(Parser):

    INTERVAL_MAP = {
        "1m": "1m",
        "5m": "5m",
        "15m": "15m",
        "30m": "30m",
        "1h": "1h",
        "4h": "4h",
        "8h": "8h",
        "1d": "1d",
        "7d": "7d",
        "1M": "30d",
    }

    def __init__(self):
        super().__init__()

    @staticmethod
    def check_response(response: dict):
        return {"code": 200, "status": "success", "data": response}

    @property
    def spot_exchange_info_parser(self) -> dict:
        return {
            "active": (lambda x: x["trade_status"] == "tradable"),
            "is_spot": True,
            "is_margin": False,
            "is_futures": False,
            "is_perp": False,
            "is_linear": True,
            "is_inverse": False,
            "symbol": (lambda x: self.parse_unified_symbol(x["base"], x["quote"])),
            "base": (lambda x: str(x["base"])),
            "quote": (lambda x: str(x["quote"])),
            "settle": (lambda x: str(x["quote"])),
            "multiplier": 1,  # spot multiplier default 1
            "leverage": 1,  # spot leverage default 1
            "listing_time": (lambda x: min([int(x["sell_start"]), int(x["buy_start"])]) * 1000),
            "expiration_time": None,  # spot not support this field
            "contract_size": 1,  # spot contract size default 1
            "tick_size": 1,
            "min_order_size": (lambda x: float(x["min_quote_amount"]) if "min_quote_amount" in x else 0),
            "max_order_size": (lambda x: float(x["max_quote_amount"]) if "max_quote_amount" in x else 0),
            "raw_data": (lambda x: x),
        }

    @property
    def perp_exchange_info_parser(self) -> dict:
        return {
            "active": True,
            "is_spot": False,
            "is_margin": False,
            "is_futures": False,
            "is_perp": True,
            "is_linear": (lambda x: True if x["settle"] in self.STABLE_CURRENCY else False),
            "is_inverse": (lambda x: False if x["settle"] in self.STABLE_CURRENCY else True),
            "symbol": (
                lambda x: self.parse_unified_symbol(
                    self.parse_perp_name(x["name"])["base"], self.parse_perp_name(x["name"])["quote"]
                )
            ),
            "base": (lambda x: self.parse_perp_name(x["name"])["base"]),
            "quote": (lambda x: self.parse_perp_name(x["name"])["quote"]),
            "settle": (lambda x: str(x["settle"])),
            "multiplier": None,
            "leverage": (lambda x: float(x["leverage_max"])),
            "listing_time": None,
            "expiration_time": None,
            "contract_size": (lambda x: self.parse_str(x["quanto_multiplier"], float)),
            "tick_size": None,
            "min_order_size": None,
            "max_order_size": None,
            "raw_data": (lambda x: x),
        }

    @property
    def futures_exchange_info_parser(self) -> dict:
        return {
            "active": (lambda x: not x["in_delisting"]),
            "is_spot": False,
            "is_margin": False,
            "is_futures": True,
            "is_perp": False,
            "is_linear": (lambda x: True if x["settle"] in self.STABLE_CURRENCY else False),
            "is_inverse": (lambda x: False if x["settle"] in self.STABLE_CURRENCY else True),
            "symbol": (
                lambda x: self.parse_unified_symbol(
                    self.parse_futures_name(x["name"])["base"], self.parse_futures_name(x["name"])["quote"]
                )
            ),
            "base": (lambda x: self.parse_base_currency(self.parse_futures_name(x["name"])["base"])),
            "quote": (lambda x: self.parse_futures_name(x["name"])["quote"]),
            "settle": (lambda x: str(x["settle"])),
            "multiplier": (lambda x: self.parse_multiplier(self.parse_futures_name(x["name"])["base"])),
            "leverage": 1,  # Not yet implemented
            "listing_time": None,
            "expiration_time": (lambda x: int(x["expire_time"]) * 1000),
            "contract_size": (lambda x: self.parse_str(x["quanto_multiplier"], float)),
            "tick_size": None,
            "min_order_size": None,
            "max_order_size": None,
            "raw_data": (lambda x: x),
        }

    def parse_perp_name(self, name: str) -> dict:
        return {
            "base": name.split("_")[0],
            "quote": name.split("_")[1],
        }

    def parse_futures_name(self, name: str) -> dict:
        return {
            "base": name.split("_")[0],
            "quote": name.split("_")[1],
            "datetime": name.split("_")[2],
        }

    def parse_exchange_info(self, response: dict, parser: dict, **kwargs) -> dict:
        response = self.check_response(response)
        datas = response["data"]

        results = {}
        for data in datas:
            data.update(kwargs)
            result = self.get_result_with_parser(data, parser)
            id = self.parse_unified_id(result)
            results[id] = result
        return results

    def get_id_map(self, exchange_info: dict, market_type: str) -> dict:
        raw_id = {
            "spot": "id",
            "futures": "name",
            "perp": "name",
        }
        infos = self.query_dict(exchange_info, {f"is_{market_type}": True})
        return {v["raw_data"][raw_id[market_type]]: k for k, v in infos.items()}

    def parse_tickers(self, response: dict, exchange_info: dict, market_type: str) -> dict:
        response = self.check_response(response)
        datas = response["data"]

        id_map = self.get_id_map(exchange_info, market_type)

        key_map = {
            "spot": "currency_pair",
            "futures": "contract",
            "perp": "contract",
        }

        results = {}
        for data in datas:
            instrument_id = id_map[data[key_map[market_type]]]
            results[instrument_id] = self.parse_ticker(data, market_type, exchange_info[instrument_id])
        return results

    def parse_ticker(self, data: dict, market_type: str, info: dict) -> dict:
        return {
            "timestamp": self.get_timestamp(),
            "instrument_id": self.parse_unified_id(info),
            "open_time": None,
            "close_time": self.get_timestamp(),
            "open": None,
            "high": self.parse_str(data["high_24h"], float),
            "low": self.parse_str(data["low_24h"], float),
            "last": self.parse_str(data["last"], float),
            "base_volume": self.parse_str(data["base_volume" if market_type == "spot" else "volume_24h_base"], float),
            "quote_volume": self.parse_str(
                data["quote_volume" if market_type == "spot" else "volume_24h_quote"], float
            ),
            "price_change": None,
            "price_change_percent": self.parse_str(data["change_percentage"], float) / 100,
            "raw_data": data,
        }

    def get_market_type(self, info: dict) -> str:
        if info["is_spot"]:
            return "spot"
        elif info["is_futures"]:
            return "futures"
        elif info["is_perp"]:
            return "perp"
        else:
            raise ValueError(f"Invalid market type. {info}")

    def get_interval(self, interval: str) -> str:
        if interval not in self.INTERVAL_MAP:
            raise ValueError(f"Invalid interval. {interval}. Must be one of {list(self.INTERVAL_MAP.keys())}")
        return self.INTERVAL_MAP[interval]

    def parse_raw_ticker(self, response: dict, market_type: str, info: dict) -> dict:
        response = self.check_response(response)
        data = response["data"][0]
        return self.parse_ticker(data, market_type, info)

    def parse_current_funding_rate(self, response: dict, info: dict) -> dict:
        response = self.check_response(response)
        data = response["data"][0]
        return {
            "timestamp": self.get_timestamp(),
            "next_funding_time": None,  # not yet implemented
            "instrument_id": self.parse_unified_id(info),
            "market_type": self.parse_unified_market_type(info),
            "funding_rate": self.parse_str(data["funding_rate"], float),
            "raw_data": data,
        }

    def parse_history_funding_rate(self, response: dict, info: dict) -> list:
        response = self.check_response(response)
        datas = response["data"]

        instrument_id = self.parse_unified_id(info)
        market_type = self.parse_unified_market_type(info)

        return [
            {
                "timestamp": self.parse_str(data["t"], int) * 1000,
                "instrument_id": instrument_id,
                "market_type": market_type,
                "funding_rate": self.parse_str(data["r"], float),
                "realized_rate": self.parse_str(data["r"], float),
                "raw_data": data,
            }
            for data in datas
        ]

    def parse_candlesticks(self, response: dict, info: dict, market_type: str, interval: str) -> any:
        response = self.check_response(response)
        datas = response["data"]
        udpate_ = {
            "instrument_id": self.parse_unified_id(info),
            "market_type": self.parse_unified_market_type(info),
            "interval": interval,
        }

        method_map = {
            "spot": self.parse_spot_candlestick,
            "futures": self.parse_futures_candlestick,
            "perp": self.parse_perp_candlestick,
        }

        results = []
        for data in datas:
            result = method_map[market_type](data, info)
            result.update(udpate_)
            results.append(result)
        return results if len(results) > 1 else results[0]

    def parse_spot_candlestick(self, data: list, info: dict) -> dict:
        return {
            "timestamp": self.parse_str(data[0], int) * 1000,
            "open": self.parse_str(data[5], float),
            "high": self.parse_str(data[3], float),
            "low": self.parse_str(data[4], float),
            "close": self.parse_str(data[2], float),
            "base_volume": self.parse_str(data[6], float),
            "quote_volume": self.parse_str(data[1], float),
            "contract_volume": self.parse_str(data[6], float),
            "raw_data": data,
        }

    def parse_perp_candlestick(self, data: dict, info: dict) -> dict:
        return {
            "timestamp": self.parse_str(data["t"], int) * 1000,
            "open": self.parse_str(data["o"], float),
            "high": self.parse_str(data["h"], float),
            "low": self.parse_str(data["l"], float),
            "close": self.parse_str(data["c"], float),
            "base_volume": self.parse_str(data["v"], float) * info["contract_size"],
            "quote_volume": self.parse_str(data["sum"], float),
            "contract_volume": self.parse_str(data["v"], float),
            "raw_data": data,
        }

    def parse_futures_candlestick(self, data: dict, info: dict) -> dict:
        return {
            "timestamp": self.parse_str(data["t"], int) * 1000,
            "open": self.parse_str(data["o"], float),
            "high": self.parse_str(data["h"], float),
            "low": self.parse_str(data["l"], float),
            "close": self.parse_str(data["c"], float),
            "base_volume": self.parse_str(data["v"], float) * info["contract_size"],
            "quote_volume": None,
            "contract_volume": self.parse_str(data["v"], float),
            "raw_data": data,
        }



from .base import Parser


class OkxParser(Parser):
    interval_map = {
        "1m": "1m",
        "3m": "3m",
        "5m": "5m",
        "15m": "15m",
        "30m": "30m",
        "1h": "1H",
        "2h": "2H",
        "4h": "4H",
        "6h": "6H",
        "12h": "12H",
        "1d": "1D",
        "2d": "2D",
        "3d": "3D",
        "1w": "1W",
        "1M": "1M",
        "3M": "3M",
    }

    market_type_map = {"spot": "SPOT", "margin": "MARGIN", "futures": "FUTURES", "perp": "SWAP"}
    _market_type_map = {"SPOT": "spot", "MARGIN": "margin", "FUTURES": "futures", "SWAP": "perp"}

    @staticmethod
    def check_response(response: dict):
        if response.get("code") == "0":
            return {"code": 200, "status": "success", "data": response["data"]}
        else:
            raise Exception(f"Error {response['code']} {response['msg']}")

    def _parse_leverage(self, lever: any):
        return int(lever) if lever else 1

    def _parse_symbol(self, data: dict) -> str:
        market_type = data["instType"]
        if market_type in ["SPOT", "MARGIN"]:
            symbol = f"{data['baseCcy']}/{data['quoteCcy']}"
        elif market_type in ["SWAP", "FUTURES"]:
            if data["ctType"] == "linear":
                symbol = f"{data['ctValCcy']}/{data['settleCcy']}"
            else:
                symbol = f"{data['settleCcy']}/{data['ctValCcy']}"
        return symbol

    @property
    def spot_margin_exchange_info_parser(self):
        return {
            "active": (lambda x: x["state"] == "live"),
            "is_spot": (lambda x: self.parse_is_spot(x["instType"])),
            "is_margin": (lambda x: self.parse_is_margin(x["instType"])),
            "is_futures": False,
            "is_perp": False,
            "is_linear": True,
            "is_inverse": False,
            "symbol": (lambda x: self._parse_symbol(x)),
            "base": (lambda x: self.parse_base_currency(x["baseCcy"])),
            "quote": (lambda x: str(x["quoteCcy"])),
            "settle": (lambda x: str(x["quoteCcy"])),
            "multiplier": 1,  # spot and margin default multiplier is 1
            "leverage": (lambda x: self._parse_leverage(x["lever"])),
            "listing_time": (lambda x: int(x["listTime"])),
            "expiration_time": None,  # spot not support this field
            "contract_size": 1,
            "tick_size": (lambda x: float(x["tickSz"])),
            "min_order_size": (lambda x: float(x["minSz"])),
            "max_order_size": (lambda x: float(x["maxMktSz"])),
            "raw_data": (lambda x: x),
        }

    @property
    def futures_perp_exchange_info_parser(self):
        return {
            "active": (lambda x: x["state"] == "live"),
            "is_spot": False,
            "is_margin": False,
            "is_futures": (lambda x: self.parse_is_futures(x["instType"])),
            "is_perp": (lambda x: self.parse_is_perpetual(x["instType"])),
            "is_linear": (lambda x: self.parse_is_linear(x["ctType"])),
            "is_inverse": (lambda x: self.parse_is_inverse(x["ctType"])),
            "symbol": (lambda x: self._parse_symbol(x)),
            "base": (lambda x: self.parse_base_currency(x["ctValCcy"] if x["ctType"] == "linear" else x["settleCcy"])),
            "quote": (lambda x: str(x["settleCcy"] if x["ctType"] == "linear" else x["ctValCcy"])),
            "settle": (lambda x: str(x["settleCcy"])),
            "multiplier": (lambda x: self.parse_multiplier(x["ctMult"])),
            "leverage": (lambda x: self._parse_leverage(x["lever"])),
            "listing_time": (lambda x: int(x["listTime"])),
            "expiration_time": (lambda x: int(x["expTime"]) if x["expTime"] else None),
            "contract_size": (
                lambda x: float(x["ctVal"])
            ),  # linear will be how many base ccy and inverse will be how many quote ccy
            "tick_size": (lambda x: float(x["tickSz"])),
            "min_order_size": (lambda x: float(x["minSz"])),
            "max_order_size": (lambda x: float(x["maxMktSz"])),
            "raw_data": (lambda x: x),
        }

    def parse_exchange_info(self, response: dict, parser: dict) -> dict:
        response = self.check_response(response)

        datas = response["data"]
        results = {}
        for data in datas:
            result = self.get_result_with_parser(data, parser)
            id = self.parse_unified_id(result)
            results[id] = result
        return results

    def combine_spot_margin_exchange_info(self, spots: dict, margins: dict) -> dict:
        for instrument_id in margins.keys():
            margin = margins[instrument_id]
            if instrument_id in spots.keys() and margin["active"]:
                spot = spots[instrument_id]

                spot["is_margin"] = True
                spot["leverage"] = margin["leverage"]

                if spot["min_order_size"] != margin["min_order_size"]:
                    print(f"min_order_size of {instrument_id} is different between spot and margin")

                if spot["max_order_size"] != margin["max_order_size"]:
                    print(f"max_order_size of {instrument_id} is different between spot and margin")

                if spot["contract_size"] != margin["contract_size"]:
                    print(f"contract_size of {instrument_id} is different between spot and margin")

                if spot["tick_size"] != margin["tick_size"]:
                    print(f"tick_size of {instrument_id} is different between spot and margin")

                spots[instrument_id] = spot
        return spots

    def parse_ticker(self, response: any, market_type: str, info: dict) -> dict:
        if "data" in response:
            response = response["data"][0]

        if market_type == "spot":
            base_volume = float(response["vol24h"])
            quote_volume = float(response["volCcy24h"])
        else:
            base_volume = float(response["volCcy24h"])
            quote_volume = float(response["volCcy24h"]) * (float(response["last"]) + float(response["open24h"])) / 2

        return {
            "timestamp": self.parse_str(response["ts"], int),
            "instrument_id": self.parse_unified_id(info),
            "open_time": None,
            "close_time": self.parse_str(response["ts"], int),
            "open": self.parse_str(response["open24h"], float),
            "high": self.parse_str(response["high24h"], float),
            "low": self.parse_str(response["low24h"], float),
            "last": self.parse_str(response["last"], float),
            "base_volume": base_volume,
            "quote_volume": quote_volume,
            "price_change": None,
            "price_change_percent": None,
            "raw_data": response,
        }

    def get_id_map(self, infos: dict, market_type: str = None) -> dict:
        if market_type:
            infos = self.query_dict(infos, {f"is_{market_type}": True})
        return {v["raw_data"]["instId"]: k for k, v in infos.items()}

    def parse_tickers(self, response: dict, market_type: str, infos: dict) -> dict:
        response = self.check_response(response)
        datas = response["data"]

        id_map = self.get_id_map(infos, market_type)
        results = {}
        for data in datas:
            instrument_id = id_map[data["instId"]]
            info = infos[instrument_id]
            results[instrument_id] = self.parse_ticker(data, market_type, info)
        return results

    def parse_funding_rates(self, response: dict, info: dict) -> list:
        response = self.check_response(response)
        datas = response["data"]

        results = []
        for data in datas:
            results.append(
                {
                    "timestamp": self.parse_str(data["fundingTime"], int),
                    "instrument_id": self.parse_unified_id(info),
                    "market_type": self.parse_unified_market_type(info),
                    "funding_rate": self.parse_str(data["fundingRate"], float),
                    "realized_rate": self.parse_str(data["realizedRate"], float),
                    "raw_data": data,
                }
            )
        return results

    def parse_current_funding_rate(self, response: dict, info: dict) -> dict:
        response = self.check_response(response)
        data = response["data"][0]
        return {
            "timestamp": self.parse_str(data["ts"], int),
            "next_funding_time": self.parse_str(data["nextFundingTime"], int),
            "instrument_id": self.parse_unified_id(info),
            "market_type": self._market_type_map[data["instType"]],
            "funding_rate": self.parse_str(data["fundingRate"], float),
            "raw_data": data,
        }

    def parse_balance(self, response: dict) -> dict:
        response = self.check_response(response)

        datas = response["data"][0]["details"]
        results = {}
        for data in datas:
            currency = data["ccy"]
            results[currency] = {
                "balance": float(data["cashBal"]),
                "available_balance": float(data["availBal"]),
                "raw_data": data,
            }
        return results

    def parse_positions(self, response: dict, infos: dict) -> dict:
        response = self.check_response(response)
        id_map = self.get_id_map(infos)
        datas = response["data"]

        results = {}
        for data in datas:
            instrument_id = id_map[data["instId"]]
            results[instrument_id] = {"position": float(data["pos"]), "raw_data": data}
        return results

    def parse_account_config(self, response: dict) -> dict:
        response = self.check_response(response)
        data = response["data"][0]

        return {
            "account_id": data["uid"],
            "main_account_id": data["mainUid"],
            "key_name": data["label"],
            "permission": data["perm"].split(","),
            "raw_data": data,
        }

    def get_interval(self, interval: str) -> str:
        if interval not in self.interval_map:
            raise ValueError(f"{interval} is not supported. Must be one of {list(self.interval_map.keys())}")
        return self.interval_map[interval]

    def parse_order_id(self, response: dict) -> str:
        response = self.check_response(response)
        data = response["data"][0]
        return str(data["ordId"])

    def parse_order_info(self, response: dict, info: dict) -> dict:
        response = self.check_response(response)
        data = response["data"][0]
        return {
            "timestamp": int(data["cTime"]),
            "instrument_id": self.parse_unified_id(info),
            "side": data["side"],
            "price": self.parse_str(data["px"], float),
            "volume": self.parse_str(data["sz"], float),
            "fee_ccy": data["feeCcy"],
            "fee": self.parse_str(data["fee"], float),
            "order_id": str(data["ordId"]),
            "order_type": data["ordType"],
            "status": data["state"],
            "raw_data": data,
        }

    def parse_cancel_order(self, response: dict) -> dict:
        response = self.check_response(response)
        data = response["data"][0]
        return {
            "order_id": str(data["ordId"]),
            "raw_data": data,
        }

    def parse_opened_orders(self, response: dict, infos: dict) -> list:
        response = self.check_response(response)
        datas = response["data"]

        results = []

        id_map = self.get_id_map(infos)

        for data in datas:
            results.append(
                {
                    "timestamp": int(data["cTime"]),
                    "instrument_id": id_map[data["instId"]],
                    "market_type": self._market_type_map[data["instType"]],
                    "side": data["side"],
                    "price": self.parse_str(data["px"], float),
                    "volume": self.parse_str(data["sz"], float),
                    "order_id": str(data["ordId"]),
                    "order_type": data["ordType"],
                    "status": data["state"],
                    "raw_data": data,
                }
            )
        return results

    def parse_history_orders(self, response: dict, infos: dict) -> list:
        response = self.check_response(response)
        datas = response["data"]

        id_map = self.get_id_map(infos)

        results = []
        for data in datas:
            results.append(
                {
                    "timestamp": int(data["fillTime"]),
                    "instrument_id": id_map[data["instId"]],
                    "market_type": self._market_type_map[data["instType"]],
                    "side": data["side"],
                    "executed_price": self.parse_str(data["fillPx"], float),
                    "executed_volume": self.parse_str(data["fillSz"], float),
                    "target_price": self.parse_str(data["px"], float),
                    "target_volume": self.parse_str(data["sz"], float),
                    "fee_currency": data["feeCcy"],
                    "fee": self.parse_str(data["fee"], float),
                    "order_type": data["ordType"],
                    "order_id": str(data["ordId"]),
                    "status": data["state"],
                    "raw_data": data,
                }
            )
        return results

    def parse_last_price(self, response: dict, instrument_id: str) -> dict:
        response = self.check_response(response)
        data = response["data"][0]
        return {
            "timestamp": self.parse_str(data["ts"], int),
            "instrument_id": instrument_id,
            "last_price": self.parse_str(data["last"], float),
            "raw_data": data,
        }

    def parse_index_price(self, response: dict, instrument_id: str) -> dict:
        response = self.check_response(response)
        data = response["data"][0]
        return {
            "timestamp": self.parse_str(data["ts"], int),
            "instrument_id": instrument_id,
            "index_price": self.parse_str(data["idxPx"], float),
            "raw_data": data,
        }

    def parse_mark_price(self, response: dict, instrument_id: str) -> dict:
        response = self.check_response(response)
        data = response["data"][0]
        return {
            "timestamp": self.parse_str(data["ts"], int),
            "instrument_id": instrument_id,
            "mark_price": self.parse_str(data["markPx"], float),
            "raw_data": data,
        }

    def parse_open_interest(self, response: dict, infos: dict) -> any:
        response = self.check_response(response)
        datas = response["data"]

        id_map = self.get_id_map(infos)

        results = []
        for data in datas:
            results.append(
                {
                    "timestamp": self.parse_str(data["ts"], int),
                    "instrument_id": id_map[data["instId"]],
                    "market_type": self._market_type_map[data["instType"]],
                    "oi_contract": self.parse_str(data["oi"], float),
                    "oi_currency": self.parse_str(data["oiCcy"], float),
                    "raw_data": data,
                }
            )

        return results[0] if len(results) == 1 else results

    def parse_orderbook(self, response: dict, info: dict) -> dict:
        response = self.check_response(response)
        datas = response["data"][0]

        asks = datas["asks"]
        bids = datas["bids"]
        return {
            "timestamp": self.parse_str(datas["ts"], int),
            "instrument_id": self.parse_unified_id(info),
            "asks": [
                {
                    "price": self.parse_str(ask[0], float),
                    "volume": self.parse_str(ask[1], float),
                    "order_number": self.parse_str(ask[3], int),
                }
                for ask in asks
            ],
            "bids": [
                {
                    "price": self.parse_str(bid[0], float),
                    "volume": self.parse_str(bid[1], float),
                    "order_number": self.parse_str(bid[3], int),
                }
                for bid in bids
            ],
            "raw_data": datas,
        }

    def parse_candlesticks(self, response: dict, info: dict, interval: str) -> any:
        def parse_volumes(data: list, market_type: str) -> dict:
            return {
                "base_volume": self.parse_str(data[5 if market_type == "spot" else 6], float),
                "quote_volume": self.parse_str(data[7], float),
                "contract_volume": self.parse_str(data[5], float),
            }

        response = self.check_response(response)
        datas = response["data"]

        instrument_id = self.parse_unified_id(info)
        market_type = self.parse_unified_market_type(info)

        results = []
        for data in datas:
            volumes = parse_volumes(data, market_type)
            results.append(
                {
                    "timestamp": self.parse_str(data[0], int),
                    "instrument_id": instrument_id,
                    "market_type": market_type,
                    "interval": interval,
                    "open": self.parse_str(data[1], float),
                    "high": self.parse_str(data[2], float),
                    "low": self.parse_str(data[3], float),
                    "close": self.parse_str(data[4], float),
                    "base_volume": volumes["base_volume"],
                    "quote_volume": volumes["quote_volume"],
                    "contract_volume": volumes["contract_volume"],
                    "raw_data": data,
                }
            )

        return results if len(results) > 1 else results[0]



import tracemalloc
from typing import Literal, Optional

from .exchanges.binance import BinanceInverse, BinanceLinear, BinanceSpot
from .parsers.binance import BinanceParser

tracemalloc.start()


class Binance(object):
    name = "binance"

    def __init__(self, api_key: str = None, api_secret: str = None):
        self.spot = BinanceSpot(api_key=api_key, api_secret=api_secret)
        self.linear = BinanceLinear()
        self.inverse = BinanceInverse()
        self.parser = BinanceParser()

        self.exchange_info = {}

    async def close(self):
        await self.spot.close()
        await self.linear.close()
        await self.inverse.close()

    async def sync_exchange_info(self) -> None:
        self.exchange_info = await self.get_exchange_info()

    async def get_exchange_info(self, market_type: Optional[Literal["spot", "margin", "futures", "perp"]] = None):
        spot = self.parser.parse_exchange_info(
            await self.spot._get_exchange_info(), self.parser.spot_exchange_info_parser
        )
        linear = self.parser.parse_exchange_info(
            await self.linear._get_exchange_info(), self.parser.futures_exchange_info_parser("linear")
        )
        inverse = self.parser.parse_exchange_info(
            await self.inverse._get_exchange_info(), self.parser.futures_exchange_info_parser("inverse")
        )
        result = {**spot, **linear, **inverse}
        return result if not market_type else self.parser.query_dict(result, {f"is_{market_type}": True})

    async def get_ticker(self, instrument_id: str):
        _symbol = self.exchange_info[instrument_id]["raw_data"]["symbol"]
        info = self.exchange_info[instrument_id]
        market_type = self.parser.get_market_type(info)

        if market_type == "spot":
            return {instrument_id: self.parser.parse_ticker(await self.spot._get_ticker(_symbol), info)}
        elif market_type == "linear":
            return {instrument_id: self.parser.parse_ticker(await self.linear._get_ticker(_symbol), info)}
        elif market_type == "inverse":
            return {instrument_id: self.parser.parse_ticker(await self.inverse._get_ticker(_symbol), info)}

    async def get_tickers(self, market_type: Optional[Literal["spot", "margin", "futures", "perp"]] = None) -> dict:
        results = {}

        tickers = [(self.spot, "spot"), (self.linear, "linear"), (self.inverse, "inverse")]

        for exchange, _market_type in tickers:
            parsed_tickers = self.parser.parse_tickers(await exchange._get_tickers(), _market_type, self.exchange_info)
            results.update(parsed_tickers)

        if market_type:
            ids = list(self.parser.query_dict(self.exchange_info, {f"is_{market_type}": True}).keys())
            return self.parser.query_dict_by_keys(results, ids)
        else:
            return results

    async def get_current_candlestick(self, instrument_id: str, interval: str) -> dict:
        if instrument_id not in self.exchange_info:
            raise ValueError(f"{instrument_id} not found in {self.name} exchange info")

        info = self.exchange_info[instrument_id]
        _symbol = info["raw_data"]["symbol"]
        _interval = self.parser.get_interval(interval)
        limit = 1
        market_type = self.parser.get_market_type(info)

        params = {
            "symbol": _symbol,
            "interval": _interval,
            "limit": limit,
        }

        method_map = {
            "spot": self.spot._get_klines,
            "linear": self.linear._get_klines,
            "inverse": self.inverse._get_klines,
        }

        return {
            instrument_id: self.parser.parse_candlesticks(
                await method_map[market_type](**params), info, market_type, interval
            )
        }

    async def get_history_candlesticks(
        self, instrument_id: str, interval: str, start: int = None, end: int = None, num: int = 500
    ) -> list:
        if instrument_id not in self.exchange_info:
            raise ValueError(f"{instrument_id} not found in {self.name} exchange info")

        info = self.exchange_info[instrument_id]
        _symbol = info["raw_data"]["symbol"]
        _interval = self.parser.get_interval(interval)
        limit = 1000
        market_type = self.parser.get_market_type(info)

        params = {
            "symbol": _symbol,
            "interval": _interval,
            "limit": limit,
        }

        method_map = {
            "spot": self.spot._get_klines,
            "linear": self.linear._get_klines,
            "inverse": self.inverse._get_klines,
        }

        query_end = None

        results = []
        if start and end:
            query_end = end
            while True:
                params["endTime"] = query_end
                result = self.parser.parse_candlesticks(
                    await method_map[market_type](**params), info, market_type, interval
                )
                results.extend(result)

                # exclude the datas with same timestamp
                results = list({v["timestamp"]: v for v in results}.values())

                query_end = min([v["timestamp"] for v in result]) - 1
                if len(result) < limit or query_end <= start:
                    break
                continue
            return sorted(
                [v for v in results if end >= v["timestamp"] >= start], key=lambda x: x["timestamp"], reverse=False
            )

        elif num:
            while True:
                params.update({"endTime": query_end} if query_end else {})
                result = self.parser.parse_candlesticks(
                    await method_map[market_type](**params), info, market_type, interval
                )

                results.extend(result)
                results = list({v["timestamp"]: v for v in results}.values())

                if len(result) < limit or len(results) >= num:
                    break

                query_end = min([v["timestamp"] for v in result]) - 1
                continue

            return sorted(results, key=lambda x: x["timestamp"], reverse=False)[-num:]

    async def get_current_funding_rate(self, instrument_id: str) -> dict:
        if instrument_id not in self.exchange_info:
            raise ValueError(f"{instrument_id} not found in {self.name} exchange info")

        info = self.exchange_info[instrument_id]
        market_type = self.parser.get_market_type(info)
        symbol = info["raw_data"]["symbol"]

        method_map = {
            "linear": self.linear._get_index_and_mark_price,
            "inverse": self.inverse._get_index_and_mark_price,
        }

        params = {"symbol": symbol}

        return {instrument_id: self.parser.parse_current_funding_rate(await method_map[market_type](**params), info)}

    async def get_history_funding_rate(
        self, instrument_id: str, start: int = None, end: int = None, num: int = 30
    ) -> list:
        if instrument_id not in self.exchange_info:
            raise ValueError(f"{instrument_id} not found in exchange info")

        info = self.exchange_info[instrument_id]
        market_type = self.parser.get_market_type(info)
        limit = 1000
        symbol = info["raw_data"]["symbol"]

        method_map = {
            "linear": self.linear._get_funding_rate_history,
            "inverse": self.inverse._get_funding_rate_history,
        }

        params = {
            "symbol": symbol,
            "limit": limit,
        }

        results = []
        query_end = None
        if start and end:
            query_start = start - 1
            query_end = end + 1
            while True:
                params.update({"startTime": query_start, "endTime": query_end})
                result = self.parser.parse_history_funding_rate(await method_map[market_type](**params), info)
                results.extend(result)
                results = list({v["timestamp"]: v for v in results}.values())

                if len(result) < limit:
                    break

                query_start = max([v["timestamp"] for v in result]) - 1

                if query_start < end:
                    break

                continue
            return sorted(
                [v for v in results if end >= v["timestamp"] >= start], key=lambda x: x["timestamp"], reverse=False
            )

        elif num:
            while True:
                params.update({"endTime": query_end} if query_end else {})
                result = self.parser.parse_history_funding_rate(await method_map[market_type](**params), info)
                results.extend(result)

                # exclude the datas with same timestamp
                results = list({v["timestamp"]: v for v in results}.values())

                if len(result) < limit or len(results) >= num:
                    break
                query_end = min([v["timestamp"] for v in result]) + 1
                continue
            return sorted(results, key=lambda x: x["timestamp"], reverse=False)[-num:]
        else:
            raise ValueError("(start, end) or num must be provided")

    async def get_last_price(self, instrument_id: str) -> dict:
        # use ticker to get last price
        if instrument_id not in self.exchange_info:
            raise ValueError(f"{instrument_id} not found in exchange info")

        return self.parser.parse_last_price(await self.get_ticker(instrument_id), instrument_id)

    async def get_index_price(self, instrument_id: str) -> dict:
        if instrument_id not in self.exchange_info:
            raise ValueError(f"{instrument_id} not found in exchange info")

        info = self.exchange_info[instrument_id]
        market_type = self.parser.get_market_type(info)
        symbol = info["raw_data"]["symbol"]

        exchange_method_map = {
            "spot": self.spot._get_margin_price_index,
            "linear": self.linear._get_index_and_mark_price,
            "inverse": self.inverse._get_index_and_mark_price,
        }

        return self.parser.parse_index_price(await exchange_method_map[market_type](symbol), info, market_type)

    async def get_mark_price(self, instrument_id: str) -> dict:
        if instrument_id not in self.exchange_info:
            raise ValueError(f"{instrument_id} not found in exchange info")

        info = self.exchange_info[instrument_id]
        market_type = self.parser.get_market_type(info)
        symbol = info["raw_data"]["symbol"]

        method_map = {
            "linear": self.linear._get_index_and_mark_price,
            "inverse": self.inverse._get_index_and_mark_price,
        }

        return self.parser.parse_mark_price(await method_map[market_type](symbol), info, market_type)

    async def get_open_interest(self, instrument_id: str) -> dict:
        if instrument_id not in self.exchange_info:
            raise ValueError(f"{instrument_id} not found in exchange info")

        info = self.exchange_info[instrument_id]
        market_type = self.parser.get_market_type(info)

        if market_type == "spot":
            raise ValueError(f"spot do not have open interest. `{instrument_id}`")

        symbol = info["raw_data"]["symbol"]
        method_map = {
            "linear": self.linear._get_open_interest,
            "inverse": self.inverse._get_open_interest,
        }

        params = {"symbol": symbol}
        return self.parser.parse_open_interest(await method_map[market_type](**params), info, market_type)

    async def get_history_open_interest(
        self, instrument_id: str, interval: str, start: int = None, end: int = None, num: int = 500
    ):
        return NotImplemented

    async def get_orderbook(self, instrument_id: str, depth: int = None) -> dict:
        if instrument_id not in self.exchange_info:
            raise ValueError(f"{instrument_id} not found in exchange info")

        info = self.exchange_info[instrument_id]
        market_type = self.parser.get_market_type(info)
        symbol = info["raw_data"]["symbol"]
        method_map = {
            "spot": self.spot._get_order_book,
            "linear": self.linear._get_order_book,
            "inverse": self.inverse._get_order_book,
        }

        limit_map = {
            "spot": 5000,
            "linear": 1000,
            "inverse": 1000,
        }

        params = {"symbol": symbol, "limit": limit_map[market_type]}
        return self.parser.parse_orderbook(await method_map[market_type](**params), info, market_type, depth=depth)

    # Private function
    async def get_spot_account_info(self) -> dict:
        return self.parser.parse_spot_account_info(await self.spot._get_account_info())

    async def get_margin_account_info(self) -> dict:
        return self.parser.parse_margin_account_info(await self.spot._get_margin_account_info())

    async def get_margin_balance(self, currency: str = None) -> dict:
        return self.parser.parse_margin_balance(await self.spot._get_margin_account_info(), currency=currency)

    async def get_margin_account_value(self, in_currency: str = None):
        info = await self.get_margin_account_info()

        net_btc_value = self.parser.parse_str(info["raw_data"]["totalNetAssetOfBtc"], float)

        instrument_id = "BTC/USDT:USDT"
        ticker = await self.get_ticker(instrument_id)
        btc_price = ticker[instrument_id]["last_price"]

        if in_currency:
            instrument_id = f"{in_currency}/USDT:USDT"
            ccy_ticker = await self.get_ticker(instrument_id)
            ccy_price = ccy_ticker[instrument_id]["last_price"]

            return net_btc_value * btc_price / ccy_price

        else:
            return net_btc_value * btc_price



from .base import Parser


class BybitParser(Parser):
    INTERVAL_MAP = {
        "1m": "1",
        "3m": "3",
        "5m": "5",
        "15m": "15",
        "30m": "30",
        "1h": "60",
        "2h": "120",
        "4h": "240",
        "6h": "360",
        "12h": "720",
        "1d": "D",
        "1M": "M",
        "1w": "W",
    }

    def check_response(self, response: dict):
        if response["retCode"] != 0:
            raise ValueError(f"Error in parsing Bybit response: {response}")
        else:
            return {
                "code": 200,
                "status": "success",
                "data": response["result"]["list"] if "list" in response["result"] else response["result"],
                "timestamp": self.parse_str(response["time"], int),
            }

    @staticmethod
    def get_market_type(info: dict) -> str:
        if info["is_spot"]:
            return "spot"
        elif info["is_linear"]:
            return "linear"
        elif info["is_inverse"]:
            return "inverse"
        else:
            raise ValueError(f"Invalid market type: {info}")

    @property
    def spot_exchange_info_parser(self) -> dict:
        return {
            "active": (lambda x: x["status"] == "Trading"),
            "is_spot": True,
            "is_margin": (lambda x: self.parse_is_margin(x["marginTrading"])),
            "is_futures": False,
            "is_perp": False,
            "is_linear": True,
            "is_inverse": False,
            "symbol": (lambda x: self.parse_unified_symbol(x["baseCoin"], x["quoteCoin"])),
            "base": (lambda x: str(x["baseCoin"])),
            "quote": (lambda x: str(x["quoteCoin"])),
            "settle": (lambda x: str(x["quoteCoin"])),
            "multiplier": 1,  # spot multiplier default 1
            "leverage": 1,  # spot leverage default 1
            "listing_time": None,  # api not support this field
            "expiration_time": None,  # spot not support this field
            "contract_size": 1,  # spot contract size default 1
            "tick_size": (lambda x: float(x["priceFilter"]["tickSize"])),
            "min_order_size": (lambda x: float(x["lotSizeFilter"]["minOrderQty"])),
            "max_order_size": (lambda x: float(x["lotSizeFilter"]["maxOrderQty"])),
            "raw_data": (lambda x: x),
        }

    @property
    def perp_futures_exchange_info_parser(self) -> dict:
        return {
            "active": (lambda x: x["status"] == "Trading"),
            "is_spot": False,
            "is_margin": False,
            "is_futures": (lambda x: self.parse_is_futures(x["contractType"])),
            "is_perp": (lambda x: self.parse_is_perpetual(x["contractType"])),
            "is_linear": (lambda x: self.parse_is_linear(x["contractType"])),
            "is_inverse": (lambda x: self.parse_is_inverse(x["contractType"])),
            "symbol": (lambda x: self.parse_unified_symbol(self.parse_base_currency(x["baseCoin"]), x["quoteCoin"])),
            "base": (lambda x: self.parse_base_currency(x["baseCoin"])),
            "quote": (lambda x: str(x["quoteCoin"])),
            "settle": (lambda x: str(x["settleCoin"])),
            "multiplier": (lambda x: self.parse_multiplier(x["baseCoin"])),
            "leverage": (lambda x: float(x["leverageFilter"]["maxLeverage"])),
            "listing_time": (lambda x: int(x["launchTime"])),
            "expiration_time": (lambda x: int(x["deliveryTime"])),
            "contract_size": (lambda x: float(x["lotSizeFilter"]["qtyStep"])),
            "tick_size": (lambda x: float(x["priceFilter"]["tickSize"])),
            "min_order_size": (lambda x: float(x["lotSizeFilter"]["minOrderQty"])),
            "max_order_size": (lambda x: float(x["lotSizeFilter"]["maxOrderQty"])),
            "raw_data": (lambda x: x),
        }

    def parse_exchange_info(self, response: dict, parser: dict) -> dict:
        response = self.check_response(response)
        datas = response["data"]

        results = {}
        for data in datas:
            result = self.get_result_with_parser(data, parser)
            id = self.parse_unified_id(result)
            results[id] = result
        return results

    def parse_tickers(self, response: dict, market_type: str, infos: dict) -> dict:
        response = self.check_response(response)
        datas = response["data"]

        id_map = self.get_id_symbol_map(infos, market_type)
        results = {}
        for data in datas:
            symbol = data["symbol"]
            if symbol not in id_map:
                print(f"Symbol {symbol} not found in Bybit exchange info.")
                continue
            instrument_id = id_map[symbol]
            results[instrument_id] = self.parse_ticker(
                data, market_type, infos[instrument_id], timestamp=response["timestamp"]
            )
        return results

    def parse_raw_ticker(self, response: dict, market_type: str, info: dict):
        response = self.check_response(response)
        data = response["data"][0]
        return self.parse_ticker(data, market_type, info, timestamp=response["timestamp"])

    def parse_ticker(self, response: dict, market_type: str, info: dict, **kwargs) -> dict:
        return {
            "timestamp": self.parse_str(kwargs["timestamp"], int),
            "instrument_id": self.parse_unified_id(info),
            "open_time": None,
            "close_time": self.parse_str(kwargs["timestamp"], int),
            "open": self.parse_str(response["prevPrice24h"], float),
            "high": self.parse_str(response["highPrice24h"], float),
            "low": self.parse_str(response["lowPrice24h"], float),
            "last": self.parse_str(response["lastPrice"], float),
            "base_volume": self.parse_str(response["volume24h" if market_type != "inverse" else "turnover24h"], float),
            "quote_volume": self.parse_str(response["turnover24h" if market_type != "inverse" else "volume24h"], float),
            "price_change": (
                self.parse_str(response["prevPrice24h"], float) - self.parse_str(response["lastPrice"], float)
            ),
            "price_change_percent": self.parse_str(response["price24hPcnt"], float),
            "raw_data": response,
        }

    def get_interval(self, interval: str) -> str:
        if interval not in self.INTERVAL_MAP:
            raise ValueError(f"Invalid interval: {interval}")
        return self.INTERVAL_MAP[interval]

    def parse_candlesticks(self, response: dict, info: dict, market_type: str, interval: str) -> any:
        response = self.check_response(response)
        datas = response["data"]

        market = self.parse_unified_market_type(info)
        instrument_id = self.parse_unified_id(info)

        results = [
            {
                "timestamp": self.parse_str(data[0], int),
                "instrument_id": instrument_id,
                "market_type": market,
                "interval": interval,
                "open": self.parse_str(data[1], float),
                "high": self.parse_str(data[2], float),
                "low": self.parse_str(data[3], float),
                "close": self.parse_str(data[4], float),
                "base_volume": self.parse_str(data[5], float),
                "quote_volume": self.parse_str(data[6], float),
                "contract_volume": (
                    self.parse_str(data[5], float) / (1 if market_type == "spot" else info["contract_size"])
                ),
                "raw_data": data,
            }
            for data in datas
        ]

        return results if len(results) > 1 else results[0]

    def get_category(self, info: dict) -> str:
        if info["is_spot"] or info["is_margin"]:
            return "spot"
        elif info["is_linear"]:
            return "linear"
        elif info["is_inverse"]:
            return "inverse"
        else:
            raise ValueError(f"Invalid market type: {info}")

    def parse_funding_rate(self, response: dict, info: dict) -> list:
        response = self.check_response(response)
        datas = response["data"]

        instrument_id = self.parse_unified_id(info)
        market_type = self.parse_unified_market_type(info)

        results = []
        for data in datas:
            results.append(
                {
                    "timestamp": self.parse_str(data["fundingRateTimestamp"], int),
                    "instrument_id": instrument_id,
                    "market_type": market_type,
                    "funding_rate": self.parse_str(data["fundingRate"], float),
                    "realized_rate": self.parse_str(data["fundingRate"], float),
                    "raw_data": data,
                }
            )
        return results

    @staticmethod
    def get_open_interest_interval(interval: str) -> str:
        interval_map = {
            "5m": "5min",
            "15m": "15min",
            "30m": "30min",
            "1h": "1h",
            "4h": "4h",
            "1d": "1d",
        }
        if interval not in interval_map:
            raise ValueError(f"Invalid interval: {interval}, should be one of {list(interval_map.keys())}")
        return interval_map[interval]

    def parse_open_interest(self, response: dict, info: dict) -> dict | list:
        response = self.check_response(response)

        results = []
        datas = response["data"]
        for datas in datas:
            results.append(
                {
                    "timestamp": self.parse_str(datas["timestamp"], int),
                    "instrument_id": self.parse_unified_id(info),
                    "open_interest": self.parse_str(datas["openInterest"], float),
                    "raw_data": datas,
                }
            )
        return results[0] if len(results) == 1 else results

    def parse_orderbook(self, response: dict, info: dict) -> dict:
        response = self.check_response(response)
        datas = response["data"]

        asks = datas["a"]
        bids = datas["b"]

        return {
            "timestamp": self.parse_str(datas["ts"], int),
            "instrument_id": self.parse_unified_id(info),
            "asks": [
                {"price": self.parse_str(ask[0], float), "volume": self.parse_str(ask[1], float), "order_number": None}
                for ask in asks
            ],
            "bids": [
                {"price": self.parse_str(bid[0], float), "volume": self.parse_str(bid[1], float), "order_number": None}
                for bid in bids
            ],
            "raw_data": datas,
        }

    def parse_last_price(self, response: dict, instrument_id: str) -> dict:
        response = self.check_response(response)
        datas = response["data"][0]
        return {
            "timestamp": response["timestamp"],
            "instrument_id": instrument_id,
            "last_price": float(datas["lastPrice"]),
            "raw_data": datas,
        }

    def parse_index_price(self, response: dict, instrument_id: str) -> dict:
        response = self.check_response(response)
        datas = response["data"][0]
        return {
            "timestamp": response["timestamp"],
            "instrument_id": instrument_id,
            "index_price": float(datas["indexPrice"]),
            "raw_data": datas,
        }

    def parse_mark_price(self, response: dict, instrument_id: str) -> dict:
        response = self.check_response(response)
        datas = response["data"][0]
        return {
            "timestamp": response["timestamp"],
            "instrument_id": instrument_id,
            "mark_price": float(datas["markPrice"]),
            "raw_data": datas,
        }

    def parse_current_funding_rate(self, response: dict, info: dict) -> dict:
        response = self.check_response(response)
        data = response["data"][0]
        return {
            "timestamp": response["timestamp"],
            "next_funding_time": self.parse_str(data["nextFundingTime"], int),
            "instrument_id": self.parse_unified_id(info),
            "market_type": self.parse_unified_market_type(info),
            "funding_rate": self.parse_str(data["fundingRate"], float),
            "raw_data": data,
        }



from .exchanges.gateio import GateioUnified
from .parsers.gateio import GateioParser


class Gateio(GateioUnified):
    name = "gateio"

    PERP_SETTLE = ["btc", "usdt", "usd"]

    def __init__(self):
        super().__init__()
        self.parser = GateioParser()
        self.exchange_info = {}

    async def sync_exchange_info(self):
        self.exchange_info = await self.get_exchange_info()

    async def get_exchange_info(self):
        spot = self.parser.parse_exchange_info(await self._get_currency_pairs(), self.parser.spot_exchange_info_parser)

        perps = {}
        for settle in self.PERP_SETTLE:
            perp = self.parser.parse_exchange_info(
                await self._get_perp_info(settle),
                self.parser.perp_exchange_info_parser,
                settle=settle.upper(),
            )
            perps.update(perp)

        settle = "usdt"
        futures = self.parser.parse_exchange_info(
            await self._get_futures_info(settle),
            self.parser.futures_exchange_info_parser,
            settle=settle.upper(),
        )

        return {**spot, **perps, **futures}

    async def get_tickers(self, market_type: str = None) -> dict:
        if market_type == "spot":
            return self.parser.parse_tickers(await self._get_spot_tickers(), self.exchange_info, "spot")
        elif market_type == "futures":
            return self.parser.parse_tickers(
                await self._get_futures_tickers(settle="usdt"), self.exchange_info, "futures"
            )
        elif market_type == "perp":
            perps = {}
            for settle in self.PERP_SETTLE:
                perp = self.parser.parse_tickers(await self._get_perp_tickers(settle), self.exchange_info, "perp")
                perps.update(perp)
            return perps
        else:
            spot = self.parser.parse_tickers(await self._get_spot_tickers(), self.exchange_info, "spot")
            futures = self.parser.parse_tickers(
                await self._get_futures_tickers(settle="usdt"), self.exchange_info, "futures"
            )
            perps = {}
            for settle in self.PERP_SETTLE:
                perp = self.parser.parse_tickers(await self._get_perp_tickers(settle), self.exchange_info, "perp")
                perps.update(perp)
            return {**spot, **futures, **perps}

    async def get_ticker(self, instrument_id: str) -> dict:
        if instrument_id not in self.exchange_info:
            raise ValueError(f"{instrument_id} not found in {self.name} exchange info")
        info = self.exchange_info[instrument_id]
        market_type = self.parser.get_market_type(info)
        method_map = {
            "spot": self._get_spot_tickers,
            "futures": self._get_futures_tickers,
            "perp": self._get_perp_tickers,
        }
        # prepare request params
        if market_type == "spot":
            params = {"currency_pair": info["raw_data"]["id"]}
        elif market_type == "futures":
            params = {"contract": info["raw_data"]["name"], "settle": info["settle"].lower()}
        else:  # perp
            params = {"contract": info["raw_data"]["name"], "settle": info["settle"].lower()}

        return {instrument_id: self.parser.parse_raw_ticker(await method_map[market_type](**params), market_type, info)}

    async def get_current_candlestick(self, instrument_id: str, interval: str) -> dict:
        if instrument_id not in self.exchange_info:
            raise ValueError(f"{instrument_id} not found in {self.name} exchange info")

        info = self.exchange_info[instrument_id]
        market_type = self.parser.get_market_type(info)
        _interval = self.parser.get_interval(interval)

        method_map = {
            "spot": self._get_spot_klines,
            "futures": self._get_futures_klines,
            "perp": self._get_perp_klines,
        }
        params = {
            "symbol": info["raw_data"]["id" if market_type == "spot" else "name"],
            "interval": _interval,
            "limit": 1,
        }

        if market_type in ["futures", "perp"]:
            params["settle"] = info["settle"].lower()

        return {
            instrument_id: self.parser.parse_candlesticks(
                await method_map[market_type](**params), info, market_type, interval
            )
        }

    async def get_history_candlesticks(
        self, instrument_id: str, interval: str, start: int = None, end: int = None, num: int = None
    ) -> list:
        if instrument_id not in self.exchange_info:
            raise ValueError(f"{instrument_id} not found in exchange_info")

        info = self.exchange_info[instrument_id]
        market_type = self.parser.get_market_type(info)
        _interval = self.parser.get_interval(interval)
        limit_map = {"spot": 1000, "futures": 2000, "perp": 2000}

        symbol_map = {
            "spot": "id",
            "futures": "name",
            "perp": "name",
        }

        method_map = {
            "spot": self._get_spot_klines,
            "futures": self._get_futures_klines,
            "perp": self._get_perp_klines,
        }

        params = {
            "symbol": info["raw_data"][symbol_map[market_type]],
            "interval": _interval,
            "limit": limit_map[market_type],
        }

        # handle special market type params
        if market_type in ["futures", "perp"]:
            params["settle"] = info["settle"].lower()

        results = []
        query_end = None
        if start and end:
            query_end = str(int(str(end)[:10]) + 1)
            while True:
                params.update({"end": query_end})
                result = self.parser.parse_candlesticks(
                    await method_map[market_type](**params), info, market_type, interval
                )
                results.extend(result)

                if len(result) < limit_map[market_type]:
                    break

                query_end = min([v["timestamp"] for v in result])

                if query_end < start:
                    break

                query_end = str(int(str(query_end)[:10]) + 1)
                continue

            return sorted(
                [v for v in results if start <= v["timestamp"] <= end], key=lambda x: x["timestamp"], reverse=False
            )

        elif num:
            while True:
                params.update({"end": query_end} if query_end else {})
                result = self.parser.parse_candlesticks(
                    await method_map[market_type](**params), info, market_type, interval
                )
                results.extend(result)

                if len(result) < limit_map[market_type] or len(results) >= num:
                    break
                query_end = str(int(str(min([v["timestamp"] for v in result]))[:10]) + 1)

                continue
            return sorted(results, key=lambda x: x["timestamp"], reverse=False)[-num:]
        else:
            raise ValueError("(start, end) or num must be provided")

    async def get_current_funding_rate(self, instrument_id: str) -> dict:
        if instrument_id not in self.exchange_info:
            raise ValueError(f"{instrument_id} not in {self.name} exchange info")

        info = self.exchange_info[instrument_id]
        market_type = self.parser.get_market_type(info)

        method_map = {
            "futures": self._get_futures_tickers,
            "perp": self._get_perp_tickers,
        }

        params = {
            "contract": info["raw_data"]["name"],
            "settle": info["settle"].lower(),
        }
        return {instrument_id: self.parser.parse_current_funding_rate(await method_map[market_type](**params), info)}

    async def get_history_funding_rate(
        self, instrument_id: str, start: int = None, end: int = None, num: int = None
    ) -> list:
        if instrument_id not in self.exchange_info:
            raise ValueError(f"{instrument_id} not in {self.name} exchange info")

        info = self.exchange_info[instrument_id]

        params = {"contract": info["raw_data"]["name"], "settle": info["settle"].lower(), "limit": 1000}

        results = self.parser.parse_history_funding_rate(await self._get_futures_funding_rate_history(**params), info)

        if start and end:
            return sorted(
                [v for v in results if start <= v["timestamp"] <= end], key=lambda x: x["timestamp"], reverse=False
            )
        elif num:
            return sorted(results, key=lambda x: x["timestamp"], reverse=False)[-num:]
        else:
            raise ValueError("(start, end) or num must be provided")



from .base import BaseClient


class BitgetUnified(BaseClient):
    BASE_URL = "https://api.bitget.com"

    def __init__(self) -> None:
        super().__init__()
        self.base_endpoint = self.BASE_URL

    async def _get_spot_exchange_info(self):
        return await self._get(self.base_endpoint + "/api/v2/spot/public/symbols")

    async def _get_derivative_exchange_info(self, product_type: str):
        params = {"productType": product_type}
        return await self._get(self.base_endpoint + "/api/v2/mix/market/contracts", params=params)

    async def _get_spot_tickers(self, symbol: str = None):
        params = {k: v for k, v in {"symbol": symbol}.items() if v}
        return await self._get(self.base_endpoint + "/api/v2/spot/market/tickers", params=params)

    async def _get_derivative_tickers(self, productType: str):
        params = {"productType": productType}
        return await self._get(self.base_endpoint + "/api/v2/mix/market/tickers", params=params)

    async def _get_derivative_ticker(self, symbol: str, productType: str):
        params = {"symbol": symbol, "productType": productType}
        return await self._get(self.base_endpoint + "/api/v2/mix/market/ticker", params=params)

    async def _get_spot_candlesticks(
        self, symbol: str, granularity: str, startTime: str = None, endTime: str = None, limit: str = None
    ):
        params = {"symbol": symbol, "granularity": granularity}
        if startTime:
            params["startTime"] = startTime
        if endTime:
            params["endTime"] = endTime
        if limit:
            params["limit"] = limit
        return await self._get(self.base_endpoint + "/api/v2/spot/market/candles", params=params)

    async def _get_derivative_candlesticks(
        self,
        symbol: str,
        productType: str,
        granularity: str,
        startTime: str = None,
        endTime: str = None,
        kLineType: str = "MARKET",
        limit: str = None,
    ):
        params = {"symbol": symbol, "productType": productType, "granularity": granularity, "kLineType": kLineType}
        if startTime:
            params["startTime"] = startTime
        if endTime:
            params["endTime"] = endTime
        if limit:
            params["limit"] = limit
        return await self._get(self.base_endpoint + "/api/v2/mix/market/candles", params=params)

    async def _get_derivative_mark_index_price(self, symbol: str, productType: str):
        params = {"symbol": symbol, "productType": productType}
        return await self._get(self.base_endpoint + "/api/v2/mix/market/symbol-price", params=params)

    async def _get_spot_orderbook(self, symbol: str, type: str = None, limit: int = None):
        params = {k: v for k, v in {"symbol": symbol, "type": type, "limit": limit}.items() if v}
        return await self._get(self.base_endpoint + "/api/v2/spot/market/orderbook", params=params)

    async def _get_spot_merge_depth(self, symbol: str, precision: str = None, limit: str = None):
        params = {k: v for k, v in {"symbol": symbol, "precision": precision, "limit": limit}.items() if v}
        return await self._get(self.base_endpoint + "/api/v2/spot/market/merge-depth", params=params)

    async def _get_derivative_merge_market_depth(
        self, symbol: str, productType: str, precision: str = None, limit: str = None
    ):
        params = {
            k: v
            for k, v in {"symbol": symbol, "productType": productType, "precision": precision, "limit": limit}.items()
            if v
        }
        return await self._get(self.base_endpoint + "/api/v2/mix/market/merge-depth", params=params)

    async def _get_derivative_current_funding_rate(self, symbol: str, productType: str):
        params = {"symbol": symbol, "productType": productType}
        return await self._get(self.base_endpoint + "/api/v2/mix/market/current-fund-rate", params=params)

    async def _get_derivative_history_funding_rate(
        self, symbol: str, productType: str, pageSize: str = None, pageNo: str = None
    ):
        params = {
            k: v
            for k, v in {"symbol": symbol, "productType": productType, "pageSize": pageSize, "pageNo": pageNo}.items()
            if v
        }
        return await self._get(self.base_endpoint + "/api/v2/mix/market/history-fund-rate", params=params)



import asyncio
import time

from .exchanges.kucoin import KucoinFutures, KucoinSpot
from .parsers.kucoin import KucoinParser
from .utils import query_dict


class Kucoin(object):
    name = "kucoin"

    def __init__(self):
        self.spot = KucoinSpot()
        self.futures = KucoinFutures()
        self.parser = KucoinParser()

        self.exchange_info = {}

    async def close(self):
        await self.spot.close()
        await self.futures.close()

    async def sync_exchange_info(self):
        self.exchange_info = await self.get_exchange_info()

    async def get_exchange_info(self) -> dict:
        spot = self.parser.parse_exchange_info(
            await self.spot._get_symbol_list(), self.parser.spot_exchange_info_parser
        )
        futures = self.parser.parse_exchange_info(
            await self.futures._get_symbol_list(), self.parser.futures_exchange_info_parser
        )

        return {**spot, **futures}

    async def get_tickers(self, market_type: str = None) -> dict:
        async def _get_derivative_tickers():
            ids = list(query_dict(self.exchange_info, "is_futures == True or is_perp == True").keys())
            num_batch = 30
            results = {}
            for i in range(0, len(ids), num_batch):
                tasks = []
                for instrument_id in ids[i : i + num_batch]:
                    _symbol = self.exchange_info[instrument_id]["raw_data"]["symbol"]
                    tasks.append(self.futures._get_symbol_detail(_symbol))
                raw_tickers = await asyncio.gather(*tasks)
                # rest for 5 sec for every 5 requests to avoid rate limit
                if i % 5 == 0:
                    time.sleep(3)
                parsed_tickers = self.parser.parse_derivative_tickers(raw_tickers, self.exchange_info)
                results.update(parsed_tickers)
            return results

        if market_type == "spot":
            return self.parser.parse_spot_tickers(await self.spot._get_tickers(), self.exchange_info)
        else:

            spot_tickers = self.parser.parse_spot_tickers(await self.spot._get_tickers(), self.exchange_info)
            derivative_tickers = await _get_derivative_tickers()
            tickers = {**spot_tickers, **derivative_tickers}

            if market_type:
                ids = list(query_dict(self.exchange_info, f"is_{market_type} == True").keys())
                return {k: v for k, v in tickers.items() if k in ids}
            else:
                return tickers

    async def get_ticker(self, instrument_id: str) -> dict:
        if instrument_id not in self.exchange_info:
            raise ValueError(f"{instrument_id} is not found in {self.name} exchange info.")

        info = self.exchange_info[instrument_id]
        market_type = "spot" if info["is_spot"] else "derivative"
        _symbol = info["raw_data"]["symbol"]

        method_map = {
            "spot": self.spot._get_24hr_stats,
            "derivative": self.futures._get_symbol_detail,
        }

        return {instrument_id: self.parser.parse_ticker(await method_map[market_type](_symbol), info, market_type)}

    async def get_last_price(self, instrument_id: str) -> dict:
        ticker = await self.get_ticker(instrument_id)
        return {
            "timestamp": ticker["timestamp"],
            "instrument_id": instrument_id,
            "last_price": ticker["last"],
            "raw_data": ticker["raw_data"],
        }

    async def get_mark_price(self, instrument_id: str) -> dict:
        if instrument_id not in self.exchange_info:
            raise ValueError(f"{instrument_id} is not found in {self.name} exchange info.")

        info = self.exchange_info[instrument_id]
        market_type = "spot" if info["is_spot"] else "derivative"
        _symbol = info["raw_data"]["symbol"]
        method_map = {
            "spot": self.spot._get_margin_mark_price,
            "derivative": self.futures._get_current_mark_price,
        }
        return self.parser.parse_mark_price(await method_map[market_type](_symbol), info, market_type)

    async def get_index_price(self, instrument_id: str):
        if instrument_id not in self.exchange_info:
            raise ValueError(f"{instrument_id} is not found in {self.name} exchange info.")

        info = self.exchange_info[instrument_id]

        if info["is_spot"]:
            raise ValueError(
                f"{instrument_id} is a spot market. Index price is not available for spot markets in `{self.name}`."
            )

        market_type = "spot" if info["is_spot"] else "derivative"
        _symbol = info["raw_data"]["symbol"]

        return self.parser.parse_index_price(await self.futures._get_current_mark_price(_symbol), info, market_type)

    async def get_orderbook(self, instrument_id: str, depth: int = None):
        if instrument_id not in self.exchange_info:
            raise ValueError(f"{instrument_id} is not found in {self.name} exchange info.")

        info = self.exchange_info[instrument_id]
        market_type = "spot" if info["is_spot"] else "derivative"
        _symbol = info["raw_data"]["symbol"]
        method_map = {
            "spot": self.spot._get_full_orderbook,
            "derivative": self.futures._get_full_orderbook,
        }

        results = self.parser.parse_orderbook(await method_map[market_type](_symbol), info, market_type)

        results["bids"] = sorted(results["bids"], key=lambda x: x["price"], reverse=True)
        results["asks"] = sorted(results["asks"], key=lambda x: x["price"])
        if depth:
            results["bids"] = results["bids"][:depth]
            results["asks"] = results["asks"][:depth]
        return results

    async def get_current_candlestick(self, instrument_id: str, interval: str) -> dict:
        if instrument_id not in self.exchange_info:
            raise ValueError(f"{instrument_id} is not found in {self.name} exchange info.")

        info = self.exchange_info[instrument_id]
        _symbol = info["raw_data"]["symbol"]
        market_type = "spot" if info["is_spot"] else "derivative"
        _interval = self.parser.get_interval(interval, market_type)

        method_map = {
            "spot": self.spot._get_klines,
            "derivative": self.futures._get_klines,
        }

        params = {
            "symbol": _symbol,
            "granularity" if market_type == "derivative" else "type": _interval,
        }

        return {
            instrument_id: self.parser.parse_current_candlestick(
                await method_map[market_type](**params), info, market_type, interval
            )
        }

    async def get_history_candlesticks(
        self, instrument_id: str, interval: str, start: int = None, end: int = None, num: int = None
    ) -> list:
        info = self.exchange_info[instrument_id]
        market_type = "spot" if info["is_spot"] else "derivative"
        _symbol = info["raw_data"]["symbol"]
        _interval = self.parser.get_interval(interval, market_type)
        method_map = {
            "spot": self.spot._get_klines,
            "derivative": self.futures._get_klines,
        }
        limit_map = {
            "spot": 100,
            "derivative": 200,
        }

        params = {
            "symbol": _symbol,
            "granularity" if market_type == "derivative" else "type": _interval,
        }

        results = []
        query_end = None
        if start and end:
            query_end = self.parser.parse_kucoin_timestamp(end, market_type) + 1
            while True:
                params.update({"end": query_end})
                klines = self.parser.parse_history_candlesticks(
                    await method_map[market_type](**params), info, market_type, interval
                )

                results.extend(klines)

                # exclude same timestamp
                results = list({v["timestamp"]: v for v in results}.values())

                if len(klines) < limit_map[market_type]:
                    break

                query_end = min([v["timestamp"] for v in results])
                if query_end < start:
                    break

                query_end = self.parser.parse_kucoin_timestamp(min([v["timestamp"] for v in results]), market_type) + 1
                continue
            return sorted(
                [v for v in results if start <= v["timestamp"] <= end], key=lambda x: x["timestamp"], reverse=False
            )

        elif num:
            while True:
                params.update({"end": query_end} if query_end else {})
                klines = self.parser.parse_history_candlesticks(
                    await method_map[market_type](**params), info, market_type, interval
                )

                results.extend(klines)

                # exclude same timestamp
                results = list({v["timestamp"]: v for v in results}.values())

                if len(results) >= num or len(klines) < limit_map[market_type]:
                    break

                query_end = (
                    int(min(v["timestamp"] for v in klines) / 1000)
                    if market_type == "spot"
                    else min(v["timestamp"] for v in klines)
                ) + 1
                continue
            return sorted(results, key=lambda x: x["timestamp"], reverse=False)[-num:]

        else:
            raise ValueError("Invalid parameters. (start, end) or (end, num) or (num) must be provided.")

    async def get_current_funding_rate(self, instrument_id: str) -> dict:
        if instrument_id not in self.exchange_info:
            raise ValueError(f"{instrument_id} is not found in {self.name} exchange info.")

        info = self.exchange_info[instrument_id]
        _symbol = info["raw_data"]["symbol"]

        return {
            instrument_id: self.parser.parse_current_funding_rate(
                await self.futures._get_current_funding_rate(_symbol), info
            )
        }

    async def get_history_funding_rate(
        self, instrument_id: str, start: int = None, end: int = None, num: int = None
    ) -> list:
        if instrument_id not in self.exchange_info:
            raise ValueError(f"{instrument_id} is not found in {self.name} exchange info.")

        info = self.exchange_info[instrument_id]
        _symbol = info["raw_data"]["symbol"]

        params = {"symbol": _symbol}

        results = []
        query_end = end if end else self.parser.get_timestamp()
        query_start = start if start else query_end - 10 * 365 * 24 * 60 * 60 * 1000
        implied_limit = 100
        if start and end:
            while True:
                params.update({"to": query_end, "_from": query_start})
                result = self.parser.parse_history_funding_rate(
                    await self.futures._get_public_funding_history(**params), info
                )
                results.extend(result)

                # exclude data with same timestamp
                results = list({v["timestamp"]: v for v in results}.values())

                if len(result) < implied_limit:
                    break

                query_end = min([v["timestamp"] for v in result])

                if min([v["timestamp"] for v in results]) < start:
                    break
                continue

            return sorted(
                [v for v in results if start <= v["timestamp"] <= end], key=lambda x: x["timestamp"], reverse=False
            )
        elif num:
            while True:
                params.update({"to": query_end, "_from": query_start})
                result = self.parser.parse_history_funding_rate(
                    await self.futures._get_public_funding_history(**params), info
                )
                results.extend(result)

                # exclude data with same timestamp
                results = list({v["timestamp"]: v for v in results}.values())

                if len(result) < implied_limit or len(results) >= num:
                    break

                query_end = min([v["timestamp"] for v in results])
                continue
            return sorted(results, key=lambda x: x["timestamp"], reverse=False)[-num:]

        else:
            raise ValueError("(start, end) or num must be provided")



from .base import BaseClient


class BinanceSpot(BaseClient):
    BASE_ENDPOINT = "https://api{}.binance.com"
    name = "binance"

    def __init__(self, api_key: str, api_secret: str, api_version: int = 3):
        super().__init__()
        self.base_endpoint = self.BASE_ENDPOINT.format(api_version)

        self.auth_data = {
            "api_key": api_key,
            "api_secret": api_secret,
        }

    async def _get_exchange_info(self):
        return await self._get(self.base_endpoint + "/api/v3/exchangeInfo")

    async def _get_ticker(self, symbol: str):
        return await self._get(self.base_endpoint + "/api/v3/ticker/24hr", params={"symbol": symbol})

    async def _get_tickers(self):
        return await self._get(self.base_endpoint + "/api/v3/ticker/24hr")

    async def _get_klines(
        self,
        symbol: str,
        interval: str,
        startTime: int = None,
        endTime: int = None,
        limit: int = 500,
        timeZone: str = "0",
    ):
        params = {"symbol": symbol, "interval": interval, "limit": limit, "timeZone": timeZone}
        if startTime:
            params["startTime"] = startTime
        if endTime:
            params["endTime"] = endTime
        return await self._get(self.base_endpoint + "/api/v3/klines", params=params)

    async def _get_orderbook(self, symbol: str, limit: int = 5000):
        params = {k: v for k, v in {"symbol": symbol, "limit": limit}.items() if v}
        return await self._get(self.base_endpoint + "/api/v3/depth", params=params)

    async def _get_margin_price_index(self, symbol: str):
        params = {"symbol": symbol}
        return await self._get(
            self.base_endpoint + "/sapi/v1/margin/priceIndex", params=params, auth_data=self.auth_data
        )

    async def _get_recent_trades_list(self, symbol: str, limit: int = 500):
        params = {k: v for k, v in {"symbol": symbol, "limit": limit}.items() if v}
        return await self._get(self.base_endpoint + "/api/v3/trades", params=params)

    # Private endpoint
    async def _get_account_info(self):
        return await self._get(self.base_endpoint + "/api/v3/account", auth_data=self.auth_data)

    async def _get_margin_account_info(self):
        return await self._get(self.base_endpoint + "/sapi/v1/margin/account", auth_data=self.auth_data)

    async def _get_order_book(self, symbol: str, limit: int = 5000):
        params = {"symbol": symbol, "limit": limit}
        return await self._get(self.base_endpoint + "/api/v3/depth", params=params)


class BinanceLinear(BaseClient):
    BASE_ENDPOINT = "https://fapi.binance.com"

    def __init__(self) -> None:
        super().__init__()
        self.linear_base_endpoint = self.BASE_ENDPOINT

    async def _get_exchange_info(self):
        return await self._get(self.linear_base_endpoint + "/fapi/v1/exchangeInfo")

    async def _get_ticker(self, symbol: str):
        return await self._get(self.linear_base_endpoint + "/fapi/v1/ticker/24hr", params={"symbol": symbol})

    async def _get_tickers(self):
        return await self._get(self.linear_base_endpoint + "/fapi/v1/ticker/24hr")

    async def _get_klines(
        self, symbol: str, interval: str, startTime: int = None, endTime: int = None, limit: int = 500
    ):
        params = {"symbol": symbol, "interval": interval, "limit": limit}
        if startTime:
            params["startTime"] = startTime
        if endTime:
            params["endTime"] = endTime
        return await self._get(self.linear_base_endpoint + "/fapi/v1/klines", params=params)

    async def _get_funding_rate_history(
        self, symbol: str, startTime: int = None, endTime: int = None, limit: int = 1000
    ):
        params = {
            k: v
            for k, v in {
                "symbol": symbol,
                "startTime": startTime,
                "endTime": endTime,
                "limit": limit,
            }.items()
            if v
        }
        return await self._get(self.linear_base_endpoint + "/fapi/v1/fundingRate", params=params)

    async def _get_index_and_mark_price(self, symbol: str):
        params = {"symbol": symbol}
        return await self._get(self.linear_base_endpoint + "/fapi/v1/premiumIndex", params=params)

    async def _get_open_interest(self, symbol: str):
        params = {"symbol": symbol}
        return await self._get(self.linear_base_endpoint + "/fapi/v1/openInterest", params=params)

    async def _get_open_interest_statistics(
        self, symbol: str, period: str, limit: int = 500, startTime: int = None, endTime: int = None
    ):
        params = {
            k: v
            for k, v in {
                "symbol": symbol,
                "period": period,
                "limit": limit,
                "startTime": startTime,
                "endTime": endTime,
            }.items()
            if v
        }
        return await self._get(self.linear_base_endpoint + "/fapi/v1/openInterestHist", params=params)

    async def _get_order_book(self, symbol: str, limit: int = 1000):
        params = {"symbol": symbol, "limit": limit}
        return await self._get(self.linear_base_endpoint + "/fapi/v1/depth", params=params)


class BinanceInverse(BaseClient):
    BASE_ENDPOINT = "https://dapi.binance.com"

    def __init__(self) -> None:
        super().__init__()
        self.inverse_base_endpoint = self.BASE_ENDPOINT

    async def _get_exchange_info(self):
        return await self._get(self.inverse_base_endpoint + "/dapi/v1/exchangeInfo")

    async def _get_ticker(self, symbol: str):
        return await self._get(self.inverse_base_endpoint + "/dapi/v1/ticker/24hr", params={"symbol": symbol})

    async def _get_tickers(self):
        return await self._get(self.inverse_base_endpoint + "/dapi/v1/ticker/24hr")

    async def _get_klines(
        self, symbol: str, interval: str, startTime: int = None, endTime: int = None, limit: int = 500
    ):
        params = {"symbol": symbol, "interval": interval, "limit": limit}
        if startTime:
            params["startTime"] = startTime
        if endTime:
            params["endTime"] = endTime
        return await self._get(self.inverse_base_endpoint + "/dapi/v1/klines", params=params)

    async def _get_funding_rate_history(
        self, symbol: str, startTime: int = None, endTime: int = None, limit: int = 1000
    ):
        params = {
            k: v
            for k, v in {
                "symbol": symbol,
                "startTime": startTime,
                "endTime": endTime,
                "limit": limit,
            }.items()
            if v
        }
        return await self._get(self.inverse_base_endpoint + "/dapi/v1/fundingRate", params=params)

    async def _get_index_and_mark_price(self, symbol: str = None, pair: str = None):
        params = {
            k: v
            for k, v in {
                "symbol": symbol,
                "pair": pair,
            }.items()
            if v
        }
        return await self._get(self.inverse_base_endpoint + "/dapi/v1/premiumIndex", params=params)

    async def _get_open_interest(self, symbol: str):
        params = {"symbol": symbol}
        return await self._get(self.inverse_base_endpoint + "/dapi/v1/openInterest", params=params)

    async def _get_open_interest_statistics(
        self, symbol: str, contractType: str, period: str, limit: int = 500, startTime: int = None, endTime: int = None
    ):
        params = {
            k: v
            for k, v in {
                "symbol": symbol,
                "contractType": contractType,
                "period": period,
                "limit": limit,
                "startTime": startTime,
                "endTime": endTime,
            }.items()
            if v
        }
        return await self._get(self.inverse_base_endpoint + "/dapi/v1/openInterestHist", params=params)

    async def _get_order_book(self, symbol: str, limit: int = 1000):
        params = {"symbol": symbol, "limit": limit}
        return await self._get(self.inverse_base_endpoint + "/dapi/v1/depth", params=params)



from ..utils import query_dict
from .base import Parser


class HtxParser(Parser):

    response_keys = ["data", "ticks"]

    INTERVAL_MAP = {
        "spot": {
            "1m": "1min",
            "5m": "5min",
            "15m": "15min",
            "30m": "30min",
            "1h": "60min",
            "4h": "4hour",
            "1d": "1day",
            "1M": "1mon",
            "1w": "1week",
            "1y": "1year",
        },
        "linear": {
            "1m": "1min",
            "5m": "5min",
            "15m": "15min",
            "30m": "30min",
            "1h": "60min",
            "4h": "4hour",
            "1d": "1day",
            "1M": "1mon",
        },
        "inverse_perp": {
            "1m": "1min",
            "5m": "5min",
            "15m": "15min",
            "30m": "30min",
            "1h": "60min",
            "4h": "4hour",
            "1d": "1day",
            "1M": "1mon",
        },
        "inverse_futures": {
            "1m": "1min",
            "5m": "5min",
            "15m": "15min",
            "30m": "30min",
            "1h": "60min",
            "4h": "4hour",
            "1d": "1day",
            "1M": "1mon",
        },
    }

    def __init__(self):
        super().__init__()

    def check_htx_response(self, response: dict):
        if response["status"] == "ok":
            for _key in self.response_keys:
                if _key in response:
                    results = {"code": 200, "status": "success", "data": response[_key]}
                    results.update({"timestamp": response["ts"]} if "ts" in response else {})
                    return results
        else:
            raise ValueError(f"Error when parsing HTX response: {response}")

    @property
    def spot_exchange_info_parser(self) -> dict:
        return {
            "active": (lambda x: x["state"] == "online"),
            "is_spot": True,
            "is_margin": False,
            "is_futures": False,
            "is_perp": False,
            "is_linear": True,
            "is_inverse": False,
            "symbol": (lambda x: self.parse_unified_symbol(x["bcdn"], x["qcdn"])),
            "base": (lambda x: self.parse_base_currency(x["bcdn"])),
            "quote": (lambda x: str(x["qcdn"])),
            "settle": (lambda x: str(x["qcdn"])),
            "multiplier": 1,  # spot and margin default multiplier is 1
            "leverage": (lambda x: float(x["lr"]) if x["lr"] else 1),  # spot and margin default leverage is 1
            "listing_time": (lambda x: int(x["toa"])),
            "expiration_time": None,  # spot not support this field
            "contract_size": 1,
            "tick_size": None,  # not yet implemented
            "min_order_size": None,  # not yet implemented
            "max_order_size": None,  # not yet implemented
            "raw_data": (lambda x: x),
        }

    @property
    def linear_exchange_info_parser(self) -> dict:
        return {
            "active": (lambda x: x["contract_status"] == 1),
            "is_spot": False,
            "is_margin": False,
            "is_futures": (lambda x: self.parse_is_futures(x["business_type"])),
            "is_perp": (lambda x: self.parse_is_perpetual(x["business_type"])),
            "is_linear": True,
            "is_inverse": False,
            "symbol": (lambda x: self.parse_unified_symbol(self.parse_pair(x)["base"], self.parse_pair(x)["quote"])),
            "base": (lambda x: self.parse_base_currency(self.parse_pair(x)["base"])),
            "quote": (lambda x: self.parse_pair(x)["quote"]),
            "settle": (lambda x: self.parse_pair(x)["quote"]),
            "multiplier": 1,
            "leverage": None,  # not yet implemented
            "listing_time": (lambda x: self.parse_str_to_timestamp(x["create_date"]) if x["create_date"] else None),
            "expiration_time": (
                lambda x: self.parse_str_to_timestamp(x["delivery_date"]) if x["delivery_date"] else None
            ),
            "contract_size": (lambda x: float(x["contract_size"])),
            "tick_size": (lambda x: float(x["price_tick"])),
            "min_order_size": None,  # not yet implemented
            "max_order_size": None,  # not yet implemented
            "raw_data": (lambda x: x),
        }

    @property
    def inverse_futures_exchange_info_parser(self):
        return {
            "active": (lambda x: x["contract_status"] == 1),
            "is_spot": False,
            "is_margin": False,
            "is_futures": True,
            "is_perp": False,
            "is_linear": False,
            "is_inverse": True,
            "symbol": (lambda x: self.parse_unified_symbol(x["symbol"], "USD")),
            "base": (lambda x: str(x["symbol"])),
            "quote": "USD",
            "settle": (lambda x: str(x["symbol"])),
            "multiplier": 1,
            "leverage": None,  # not yet implemented
            "listing_time": (lambda x: self.parse_str_to_timestamp(x["create_date"]) if x["create_date"] else None),
            "expiration_time": (lambda x: int(x["delivery_time"]) if x["delivery_time"] else None),
            "contract_size": (lambda x: float(x["contract_size"])),
            "tick_size": (lambda x: float(x["price_tick"])),
            "min_order_size": None,  # not yet implemented
            "max_order_size": None,  # not yet implemented
            "raw_data": (lambda x: x),
        }

    @property
    def inverse_perp_exchange_info_parser(self):
        return {
            "active": (lambda x: x["contract_status"] == 1),
            "is_spot": False,
            "is_margin": False,
            "is_futures": False,
            "is_perp": True,
            "is_linear": False,
            "is_inverse": True,
            "symbol": (lambda x: self.parse_unified_symbol(x["symbol"], "USD")),
            "base": (lambda x: str(x["symbol"])),
            "quote": "USD",
            "settle": (lambda x: str(x["symbol"])),
            "multiplier": 1,
            "leverage": None,  # not yet implemented
            "listing_time": (lambda x: self.parse_str_to_timestamp(x["create_date"]) if x["create_date"] else None),
            "expiration_time": None,
            "contract_size": (lambda x: float(x["contract_size"])),
            "tick_size": (lambda x: float(x["price_tick"])),
            "min_order_size": None,  # not yet implemented
            "max_order_size": None,  # not yet implemented
            "raw_data": (lambda x: x),
        }

    @staticmethod
    def parse_pair(response: dict) -> dict:
        if response["delivery_date"]:
            return {
                "base": response["pair"].split("-")[0],
                "quote": response["pair"].split("-")[1],
                "datetime": response["delivery_date"],
            }
        else:
            return {
                "base": response["pair"].split("-")[0],
                "quote": response["pair"].split("-")[1],
            }

    def parse_exchange_info(self, response: dict, parser: dict):
        response = self.check_htx_response(response)

        results = {}
        datas = response["data"]
        for data in datas:
            result = self.get_result_with_parser(data, parser)
            id = self.parse_unified_id(result)
            results[id] = result
        return results

    def get_htx_id_map(self, exchange_info: dict, market_type: str) -> dict:
        keys_map = {
            "spot": "sc",
            "linear": "contract_code",
            "inverse_perp": "contract_code",
            "inverse_futures": "symbol",
        }
        if market_type == "linear":
            infos = query_dict(exchange_info, "(is_perp == True or is_futures == True) and is_linear == True")
        elif market_type == "inverse_perp":
            infos = query_dict(exchange_info, "is_perp == True and is_inverse == True")
        elif market_type == "inverse_futures":
            infos = query_dict(exchange_info, "is_futures == True and is_inverse == True")
            return {self.parse_inverse_futures_symbol(v["raw_data"]): k for k, v in infos.items()}
        else:
            infos = self.query_dict(exchange_info, {f"is_{market_type}": True})
        return {v["raw_data"][keys_map[market_type]]: k for k, v in infos.items()}

    def parse_inverse_futures_symbol(self, datas: dict) -> str:
        contract_type_map = {
            "this_week": "CW",
            "next_week": "NW",
            "quarter": "CQ",
        }
        return f"{datas['symbol']}_{contract_type_map[datas['contract_type']]}"

    def parse_tickers(self, response: dict, exchange_infos: dict, market_type: str) -> dict:
        response = self.check_htx_response(response)

        method_map = {
            "spot": self.parse_spot_ticker,
            "linear": self.parse_linear_ticker,
            "inverse_perp": self.parse_inverse_perp_ticker,
            "inverse_futures": self.parse_inverse_futures_ticker,
        }

        keys_map = {
            "spot": "symbol",
            "linear": "contract_code",
            "inverse_perp": "contract_code",
            "inverse_futures": "symbol",
        }

        id_map = self.get_htx_id_map(exchange_infos, market_type)

        results = {}
        datas = response["data"]
        for data in datas:
            if data[keys_map[market_type]] not in id_map:
                print(f"Unmapped symbol: {data[keys_map[market_type]]} in {market_type}")
                continue
            instrument_id = id_map[data[keys_map[market_type]]]
            params = {
                "response": data,
                "info": exchange_infos[instrument_id],
            }
            if market_type == "spot":
                params["timestamp"] = response["timestamp"]
            results[instrument_id] = method_map[market_type](**params)
        return results

    def parse_ticker(self, response: dict, market_type: str, info: dict) -> dict:
        method_map = {
            "spot": self.parse_spot_ticker,
            "linear": self.parse_linear_ticker,
            "inverse_perp": self.parse_inverse_perp_ticker,
            "inverse_futures": self.parse_inverse_futures_ticker,
        }
        params = {
            "response": response,
            "info": info,
        }
        if market_type == "spot":
            params["timestamp"] = response["ts"]
        return method_map[market_type](**params)

    def parse_spot_ticker(self, response: dict, info: dict, timestamp: str):
        return {
            "timestamp": self.parse_str(timestamp, int),
            "instrument_id": self.parse_unified_id(info),
            "open_time": None,
            "close_time": None,
            "open": self.parse_str(response["open"], float),
            "high": self.parse_str(response["high"], float),
            "low": self.parse_str(response["low"], float),
            "last": self.parse_str(response["close"], float),
            "base_volume": self.parse_str(response["amount"], float),
            "quote_volume": self.parse_str(response["vol"], float),
            "price_change": self.parse_str(response["close"], float) - self.parse_str(response["open"], float),
            "price_change_percent": None,
            "raw_data": response,
        }

    def parse_linear_ticker(self, response: dict, info: dict):
        return {
            "timestamp": self.parse_str(response["ts"], int),
            "instrument_id": self.parse_unified_id(info),
            "open_time": None,
            "close_time": self.parse_str(response["ts"], int),
            "open": self.parse_str(response["open"], float),
            "high": self.parse_str(response["high"], float),
            "low": self.parse_str(response["low"], float),
            "last": self.parse_str(response["close"], float),
            "base_volume": self.parse_str(response["amount"], float) * info["contract_size"],
            "quote_volume": self.parse_str(response["vol"], float),
            "price_change": self.parse_str(response["close"], float) - self.parse_str(response["open"], float),
            "price_change_percent": None,
            "raw_data": response,
        }

    def parse_inverse_perp_ticker(self, response: dict, info: dict):
        return {
            "timestamp": self.parse_str(response["ts"], int),
            "instrument_id": self.parse_unified_id(info),
            "open_time": None,
            "close_time": self.parse_str(response["ts"], int),
            "open": self.parse_str(response["open"], float),
            "high": self.parse_str(response["high"], float),
            "low": self.parse_str(response["low"], float),
            "last": self.parse_str(response["close"], float),
            "base_volume": self.parse_str(response["amount"], float),
            "quote_volume": self.parse_str(response["vol"], float),
            "price_change": self.parse_str(response["close"], float) - self.parse_str(response["open"], float),
            "price_change_percent": None,
            "raw_data": response,
        }

    def parse_inverse_futures_ticker(self, response: dict, info: dict):
        return {
            "timestamp": self.parse_str(response["ts"], int),
            "instrument_id": self.parse_unified_id(info),
            "open_time": None,
            "close_time": self.parse_str(response["ts"], int),
            "open": self.parse_str(response["open"], float),
            "high": self.parse_str(response["high"], float),
            "low": self.parse_str(response["low"], float),
            "last": self.parse_str(response["close"], float),
            "base_volume": self.parse_str(response["amount"], float),
            "quote_volume": self.parse_str(response["vol"], float),
            "price_change": self.parse_str(response["close"], float) - self.parse_str(response["open"], float),
            "price_change_percent": None,
            "raw_data": response,
        }

    def get_market_type(self, info: dict) -> str:
        if info["is_spot"]:
            return "spot"
        elif info["is_linear"] and (info["is_futures"] or info["is_perp"]):
            return "linear"
        elif info["is_inverse"] and info["is_futures"]:
            return "inverse_futures"
        elif info["is_inverse"] and info["is_perp"]:
            return "inverse_perp"
        else:
            raise ValueError("Unknown market type")

    def get_interval(self, interval: str, market_type: str) -> str:
        if interval not in self.INTERVAL_MAP[market_type]:
            raise ValueError(
                f"Invalid interval: {interval} in {market_type}, must be one of {list(self.INTERVAL_MAP[market_type])}"
            )
        return self.INTERVAL_MAP[market_type][interval]

    def parse_current_funding_rate(self, response: dict, info: dict) -> dict:
        response = self.check_htx_response(response)
        data = response["data"]
        return {
            "timestamp": response["timestamp"],
            "next_funding_time": self.parse_str(data["funding_time"], int),
            "instrument_id": self.parse_unified_id(info),
            "market_type": self.parse_unified_market_type(info),
            "funding_rate": self.parse_str(data["funding_rate"], float),
            "raw_data": data,
        }

    def parse_history_funding_rate(self, response: dict, info: dict) -> list:
        response = self.check_htx_response(response)
        datas = response["data"]["data"]

        instrument_id = self.parse_unified_id(info)
        market_type = self.parse_unified_market_type(info)

        return [
            {
                "timestamp": self.parse_str(data["funding_time"], int),
                "instrument_id": instrument_id,
                "market_type": market_type,
                "funding_rate": self.parse_str(data["funding_rate"], float),
                "realized_rate": self.parse_str(data["realized_rate"], float),
                "raw_data": data,
            }
            for data in datas
        ]

    def parse_candlesticks(self, response: dict, info: dict, market_type: str, interval: str) -> any:
        """
        - spot :
        [
            {
                'id': 1713110400,
                'open': 64186.51,
                'close': 66184.98,
                'low': 62578.94,
                'high': 66566.0,
                'amount': 1187.8846778540208,
                'vol': 77002399.88436274,
                'count': 86506
            }
        ]

        - linear :
        {
            "amount":0.004
            "close":13076.8
            "count":1
            "high":13076.8
            "id":1603695060
            "low":13076.8
            "open":13076.8
            "trade_turnover":52.3072
            "vol":4
        }

        - inverse:
        {
            "id":1628652420
            "open":45875.02
            "close":45851
            "low":45850.93
            "high":45880.01
            "amount":2.6471133153472475
            "vol":1214
            "count":50
        }

        """

        response = self.check_htx_response(response)
        datas = response["data"]

        update_ = {
            "instrument_id": self.parse_unified_id(info),
            "market_type": self.parse_unified_market_type(info),
            "interval": interval,
        }

        results = []

        for data in datas:
            result = self.parse_candlestick(data, info, market_type)
            result.update(update_)
            results.append(result)

        return results if len(results) > 1 else results[0]

    def parse_candlestick(self, data: dict, info: dict, market_type: str) -> dict:

        return {
            "timestamp": self.parse_str(data["id"], int) * 1000,
            "open": self.parse_str(data["open"], float),
            "high": self.parse_str(data["high"], float),
            "low": self.parse_str(data["low"], float),
            "close": self.parse_str(data["close"], float),
            "base_volume": self.parse_str(data["amount"], float),
            "quote_volume": (
                self.parse_str(data["vol" if market_type != "linear" else "trade_turnover"], float)
                * (1 if info["is_linear"] else info["contract_size"])
            ),
            "contract_volume": self.parse_str(data["amount" if market_type == "spot" else "vol"], float),
            "raw_data": data,
        }



from .exchanges.bitget import BitgetUnified
from .parsers.bitget import BitgetParser
from .utils import query_dict


class Bitget(BitgetUnified):
    name = "bitget"

    def __init__(self) -> None:
        super().__init__()
        self.parser = BitgetParser()
        self.exchange_info = {}

    async def sync_exchange_info(self):
        self.exchange_info = await self.get_exchange_info()

    async def get_exchange_info(self, market_type: str = None):
        if market_type:
            if market_type == "spot":
                return self.parser.parse_exchange_info(
                    await self._get_spot_exchange_info(), self.parser.spot_exchange_info_parser
                )
            else:
                derivative = {}
                inverse = self.parser.parse_exchange_info(
                    await self._get_derivative_exchange_info("COIN-FUTURES"),
                    self.parser.derivative_exchange_info_parser,
                )
                derivative.update(inverse)

                for settle in self.parser.LINEAR_FUTURES_SETTLE:
                    sub_linear = self.parser.parse_exchange_info(
                        await self._get_derivative_exchange_info(f"{settle}-FUTURES"),
                        self.parser.derivative_exchange_info_parser,
                    )
                    derivative.update(sub_linear)

                instrument_id = list(query_dict(self.exchange_info, f"is_{market_type} == True").keys())
                return {k: v for k, v in derivative.items() if k in instrument_id}
        else:
            spot = self.parser.parse_exchange_info(
                await self._get_spot_exchange_info(), self.parser.spot_exchange_info_parser
            )

            inverse = self.parser.parse_exchange_info(
                await self._get_derivative_exchange_info("COIN-FUTURES"),
                self.parser.derivative_exchange_info_parser,
            )

            linear_futures = {}
            for settle in self.parser.LINEAR_FUTURES_SETTLE:
                sub_linear = self.parser.parse_exchange_info(
                    await self._get_derivative_exchange_info(f"{settle}-FUTURES"),
                    self.parser.derivative_exchange_info_parser,
                )
                linear_futures.update(sub_linear)

            return {**spot, **inverse, **linear_futures}

    async def get_tickers(self, market_type: str = None):
        if market_type:
            if market_type == "spot":
                return self.parser.parse_tickers(await self._get_spot_tickers(), self.exchange_info, "spot")
            else:
                derivative = {}
                inverse = self.parser.parse_tickers(
                    await self._get_derivative_tickers("COIN-FUTURES"), self.exchange_info, "derivative"
                )
                derivative.update(inverse)

                for settle in self.parser.LINEAR_FUTURES_SETTLE:
                    sub_linear = self.parser.parse_tickers(
                        await self._get_derivative_tickers(f"{settle}-FUTURES"), self.exchange_info, "derivative"
                    )
                    derivative.update(sub_linear)

                instrument_id = list(query_dict(self.exchange_info, f"is_{market_type} == True").keys())
                return {k: v for k, v in derivative.items() if k in instrument_id}
        else:
            spot = self.parser.parse_tickers(await self._get_spot_tickers(), self.exchange_info, "spot")
            inverse = self.parser.parse_tickers(
                await self._get_derivative_tickers("COIN-FUTURES"), self.exchange_info, "derivative"
            )
            linear_futures = {}
            for settle in self.parser.LINEAR_FUTURES_SETTLE:
                sub_linear = self.parser.parse_tickers(
                    await self._get_derivative_tickers(f"{settle}-FUTURES"), self.exchange_info, "derivative"
                )
                linear_futures.update(sub_linear)
            return {**spot, **inverse, **linear_futures}

    async def get_ticker(self, instrument_id: str) -> dict:
        if instrument_id not in self.exchange_info:
            raise f"{instrument_id} not found in {self.name} exchange info"

        info = self.exchange_info[instrument_id]
        _symbol = info["raw_data"]["symbol"]
        market_type = self.parser.get_market_type(info)
        product_type = self.parser.get_product_type(info)

        method_map = {
            "spot": self._get_spot_tickers,
            "derivative": self._get_derivative_ticker,
        }
        params = {"symbol": _symbol} if market_type == "spot" else {"symbol": _symbol, "productType": product_type}
        return {instrument_id: self.parser.parse_raw_ticker(await method_map[market_type](**params), info, market_type)}

    async def get_last_price(self, instrument_id: str) -> dict:
        ticker = await self.get_ticker(instrument_id)
        return {
            "timestamp": ticker["timestamp"],
            "instrument_id": instrument_id,
            "last_price": ticker["last"],
            "raw_data": ticker["raw_data"],
        }

    async def get_index_price(self, instrument_id: str) -> dict:
        if instrument_id not in self.exchange_info:
            raise f"{instrument_id} not found in {self.name} exchange info"

        info = self.exchange_info[instrument_id]
        if info["is_spot"]:
            raise ValueError(f"{instrument_id} is not a derivative instrument")
        product_type = self.parser.get_product_type(info)
        _symbol = info["raw_data"]["symbol"]
        return self.parser.parse_mark_index_price(
            await self._get_derivative_mark_index_price(_symbol, product_type), info, "index"
        )

    async def get_mark_price(self, instrument_id: str) -> dict:
        if instrument_id not in self.exchange_info:
            raise f"{instrument_id} not found in {self.name} exchange info"

        info = self.exchange_info[instrument_id]
        if info["is_spot"]:
            raise ValueError(f"{instrument_id} is not a derivative instrument")
        product_type = self.parser.get_product_type(info)
        _symbol = info["raw_data"]["symbol"]
        return self.parser.parse_mark_index_price(
            await self._get_derivative_mark_index_price(_symbol, product_type), info, "mark"
        )

    async def get_orderbook(self, instrument_id: str, depth: int = None) -> dict:
        if instrument_id not in self.exchange_info:
            raise f"{instrument_id} not found in {self.name} exchange info"

        info = self.exchange_info[instrument_id]
        _symbol = info["raw_data"]["symbol"]
        _market_type = self.parser.get_market_type(info)
        limit = "max"
        params = {
            "symbol": _symbol,
            "limit": limit,
        }

        if _market_type == "derivative":
            params.update({"productType": self.parser.get_product_type(info)})
        method_map = {
            "spot": self._get_spot_merge_depth,
            "derivative": self._get_derivative_merge_market_depth,
        }
        orderbook = self.parser.parse_orderbook(await method_map[_market_type](**params), info)

        if depth:
            orderbook["asks"] = sorted(orderbook["asks"], key=lambda x: x["price"], reverse=False)[:depth]
            orderbook["bids"] = sorted(orderbook["bids"], key=lambda x: x["price"], reverse=True)[:depth]

        return orderbook

    async def get_current_candlestick(self, instrument_id: str, interval: str) -> dict:
        if instrument_id not in self.exchange_info:
            raise ValueError(f"{instrument_id} not found in {self.name} exchange info")

        info = self.exchange_info[instrument_id]
        _symbol = info["raw_data"]["symbol"]
        market_type = self.parser.get_market_type(info)
        _interval = self.parser.get_interval(interval, market_type)
        limit = 1

        method_map = {
            "spot": self._get_spot_candlesticks,
            "derivative": self._get_derivative_candlesticks,
        }

        params = {
            "symbol": _symbol,
            "granularity": _interval,
            "limit": limit,
        }

        if market_type == "derivative":
            params.update({"productType": self.parser.get_product_type(info)})

        return {
            instrument_id: self.parser.parse_candlesticks(
                await method_map[market_type](**params), info, market_type, interval
            )
        }

    async def get_history_candlesticks(
        self, instrument_id: str, interval: str, start: int = None, end: int = None, num: int = None
    ) -> list:
        if instrument_id not in self.exchange_info:
            raise ValueError(f"{instrument_id} not found in {self.name} exchange info")

        info = self.exchange_info[instrument_id]
        market_type = self.parser.get_market_type(info)
        _symbol = info["raw_data"]["symbol"]
        _interval = self.parser.get_interval(interval, market_type)
        limit = 300

        method_map = {
            "spot": self._get_spot_candlesticks,
            "derivative": self._get_derivative_candlesticks,
        }

        params = {
            "symbol": _symbol,
            "granularity": _interval,
            "limit": limit,
        }
        if market_type == "derivative":
            params.update({"productType": self.parser.get_product_type(self.exchange_info[instrument_id])})

        results = []
        query_end = None
        if start and end:
            query_end = end
            while True:
                params.update({"endTime": query_end})
                result = self.parser.parse_candlesticks(
                    await method_map[market_type](**params), info, market_type, interval
                )
                results.extend(result)

                # exclude data with same timestamp
                results = list({v["timestamp"]: v for v in results}.values())

                if not result or len(result) < limit:
                    break

                query_end = min([v["timestamp"] for v in result]) + 1

                if query_end < start:
                    break

            return sorted(
                [i for i in results if start <= i["timestamp"] <= end], key=lambda x: x["timestamp"], reverse=False
            )
        elif num:
            while True:
                params.update({"endTime": query_end})

                result = self.parser.parse_candlesticks(
                    await method_map[market_type](**params), info, market_type, interval
                )
                results.extend(result)

                # exclude data with same timestamp
                results = list({v["timestamp"]: v for v in results}.values())

                if not result or len(result) < limit:
                    break

                query_end = min([v["timestamp"] for v in result]) + 1

                if len(results) >= num:
                    break
            return sorted(results, key=lambda x: x["timestamp"], reverse=False)[-num:]

        else:
            raise ValueError("(start, end) or num must be provided")

    async def get_current_funding_rate(self, instrument_id: str) -> dict:
        if instrument_id not in self.exchange_info:
            raise f"{instrument_id} not found in {self.name} exchange info"

        info = self.exchange_info[instrument_id]
        _symbol = info["raw_data"]["symbol"]
        _product_type = self.parser.get_product_type(info)

        return {
            instrument_id: self.parser.parse_current_funding_rate(
                await self._get_derivative_current_funding_rate(_symbol, _product_type), info
            )
        }

    async def get_history_funding_rate(
        self, instrument_id: str, start: int = None, end: int = None, num: int = 30
    ) -> list:
        if instrument_id not in self.exchange_info:
            raise f"{instrument_id} not found in {self.name} exchange info"

        info = self.exchange_info[instrument_id]
        _symbol = info["raw_data"]["symbol"]
        _product_type = self.parser.get_product_type(info)
        limit = 100
        page = 1
        params = {"symbol": _symbol, "productType": _product_type, "pageSize": limit, "pageNo": page}

        results = []
        if start and end:
            while True:
                params.update({"pageNo": page})
                result = self.parser.parse_history_funding_rate(
                    await self._get_derivative_history_funding_rate(**params), info
                )
                results.extend(result)

                # exclude data with same timestamp
                results = list({v["timestamp"]: v for v in results}.values())

                if len(result) < limit:
                    break
                min_timestamp = min([i["timestamp"] for i in result])
                if min_timestamp < start:
                    break

            return sorted(
                [i for i in results if start <= i["timestamp"] <= end], key=lambda x: x["timestamp"], reverse=False
            )
        elif num:
            while True:
                params.update({"pageNo": page})
                result = self.parser.parse_history_funding_rate(
                    await self._get_derivative_history_funding_rate(**params), info
                )
                results.extend(result)

                # exclude data with same timestamp
                results = list({v["timestamp"]: v for v in results}.values())

                if len(result) < limit or len(results) > num:
                    break
                page += 1
                continue
            return sorted(results, key=lambda x: x["timestamp"], reverse=False)[-num:]
        else:
            raise ValueError("(start, end) or num must be provided")
